import logging
import time
from functools import partial
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np

from qcodes.instrument.base import Instrument, InstrumentBase
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.instrument.parameter import (
    ArrayParameter,
    MultiParameter,
    Parameter,
    ParameterWithSetpoints,
)
from qcodes.utils.validators import Arrays, ComplexNumbers, Numbers, OnOff
from qcodes.utils.validators import Sequence as ValidatorSequence
from qcodes.utils.validators import Strings

log = logging.getLogger(__name__)


class MockParabola(Instrument):
    """
    Holds dummy parameters which are get and set able as well as provides
    some basic functions that depends on these parameters for testing
    purposes.

    This instrument is intended to be simpler than the mock model in that it
    does not emulate communications.

    It has 3 main parameters (x, y, z) in order to allow for testing of 3D
    sweeps. The function (parabola with optional noise) is chosen to allow
    testing of numerical optimizations.
    """

    def __init__(self, name, **kw):
        super().__init__(name, **kw)

        # Instrument parameters
        for parname in ['x', 'y', 'z']:
            self.add_parameter(parname, unit='a.u.',
                               parameter_class=Parameter,
                               vals=Numbers(), initial_value=0,
                               get_cmd=None, set_cmd=None)

        self.add_parameter('noise', unit='a.u.',
                           label='white noise amplitude',
                           parameter_class=Parameter,
                           vals=Numbers(), initial_value=0,
                           get_cmd=None, set_cmd=None)

        self.add_parameter('parabola', unit='a.u.',
                           get_cmd=self._measure_parabola)
        self.add_parameter('skewed_parabola', unit='a.u.',
                           get_cmd=self._measure_skewed_parabola)

    def _measure_parabola(self):
        return (self.x.get()**2 + self.y.get()**2 + self.z.get()**2 +
                self.noise.get()*np.random.rand(1))

    def _measure_skewed_parabola(self):
        """
        Adds an -x term to add a corelation between the parameters.
        """
        return ((self.x.get()**2 + self.y.get()**2 +
                 self.z.get()**2)*(1 + abs(self.y.get()-self.x.get())) +
                 self.noise.get()*np.random.rand(1))


class MockMetaParabola(InstrumentBase):
    """
    Test for a meta instrument, has a tunable gain knob

    Unlike a MockParabola, a MockMetaParabola does not have a connection, or
    access to ask_raw/write_raw, i.e. it would not be connected to a real instrument.

    It is also not tracked in the global _all_instruments list, but is still
    snapshottable in a station.
    """

    def __init__(self, name, mock_parabola_inst, **kw):
        """
        Create a new MockMetaParabola, connected to an existing MockParabola instance.
        """
        super().__init__(name, **kw)
        self.mock_parabola_inst = mock_parabola_inst

        # Instrument parameters
        for parname in ['x', 'y', 'z']:
            self.parameters[parname] = getattr(mock_parabola_inst, parname)
        self.add_parameter('gain', parameter_class=Parameter,
                           initial_value=1,
                           get_cmd=None, set_cmd=None)

        self.add_parameter('parabola', unit='a.u.',
                           get_cmd=self._get_parabola)
        self.add_parameter('skewed_parabola', unit='a.u.',
                           get_cmd=self._get_skew_parabola)

    def _get_parabola(self):
        val = self.mock_parabola_inst.parabola.get()
        return val*self.gain.get()

    def _get_skew_parabola(self):
        val = self.mock_parabola_inst.skewed_parabola.get()
        return val*self.gain.get()


class DummyInstrument(Instrument):

    def __init__(self, name: str = 'dummy',
                 gates: Sequence[str] = ('dac1', 'dac2', 'dac3'), **kwargs):

        """
        Create a dummy instrument that can be used for testing

        Args:
            name: name for the instrument
            gates: list of names that is used to create parameters for
                            the instrument
        """
        super().__init__(name, **kwargs)

        # make gates
        for _, g in enumerate(gates):
            self.add_parameter(
                g,
                parameter_class=Parameter,
                initial_value=0,
                label=f"Gate {g}",
                unit="V",
                vals=Numbers(-800, 400),
                get_cmd=None,
                set_cmd=None,
            )


class DummyAttrInstrument(Instrument):
    def __init__(self, name: str = "dummy", **kwargs: Any):

        """
        Create a dummy instrument that can be used for testing.
        This instrument has its parameters declared as attributes
        and does not use add_parameter.
        """
        super().__init__(name, **kwargs)

        self.ch1 = Parameter(
            "ch1",
            instrument=self,
            initial_value=0,
            label="Gate ch1",
            unit="V",
            vals=Numbers(-800, 400),
            get_cmd=None,
            set_cmd=None,
        )

        self.ch2 = Parameter(
            "ch2",
            instrument=self,
            initial_value=0,
            label="Gate ch2",
            unit="V",
            vals=Numbers(-800, 400),
            get_cmd=None,
            set_cmd=None,
        )


class DmmExponentialParameter(Parameter):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self._ed = self._exponential_decay(5, 0.2)
        next(self._ed)

    def get_raw(self):
        """
        This method is automatically wrapped to
        provide a ``get`` method on the parameter instance.
        """
        dac = self.root_instrument._setter_instr
        val = self._ed.send(dac.ch1())
        next(self._ed)
        return val

    @staticmethod
    def _exponential_decay(a: float, b: float):
        """
        Yields a*exp(-b*x) where x is put in
        """
        x = 0
        while True:
            x = yield
            yield a * np.exp(-b * x) + 0.02 * a * np.random.randn()


class DmmGaussParameter(Parameter):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.x0 = 0.1
        self.y0 = 0.2
        self.sigma = 0.25
        self.noise: float = 0.0005
        self._gauss = self._gauss_model()
        next(self._gauss)

    def get_raw(self):
        """
        This method is automatically wrapped to
        provide a ``get`` method on the parameter instance.
        """
        dac = self.root_instrument._setter_instr
        val = self._gauss.send((dac.ch1.get(), dac.ch2.get()))
        next(self._gauss)
        return val

    def _gauss_model(self):
        """
        Returns a generator sampling a gaussian. The gaussian is
        normalised such that its maximal value is simply 1
        """
        while True:
            (x, y) = yield
            model = np.exp(-((self.x0-x)**2+(self.y0-y)**2)/2/self.sigma**2)*np.exp(2*self.sigma**2)
            noise = np.random.randn()*self.noise
            yield model + noise


class DummyInstrumentWithMeasurement(Instrument):

    def __init__(
            self,
            name: str,
            setter_instr: DummyInstrument,
            **kwargs):
        super().__init__(name=name, **kwargs)
        self._setter_instr = setter_instr
        self.add_parameter('v1',
                           parameter_class=DmmExponentialParameter,
                           initial_value=0,
                           label='Gate v1',
                           unit="V",
                           vals=Numbers(-800, 400),
                           get_cmd=None, set_cmd=None)
        self.add_parameter('v2',
                           parameter_class=DmmGaussParameter,
                           initial_value=0,
                           label='Gate v2',
                           unit="V",
                           vals=Numbers(-800, 400),
                           get_cmd=None, set_cmd=None)


class DummyChannel(InstrumentChannel):
    """
    A single dummy channel implementation
    """

    def __init__(self, parent, name, channel):
        super().__init__(parent, name)

        self._channel = channel

        # Add the various channel parameters
        self.add_parameter('temperature',
                           parameter_class=Parameter,
                           initial_value=0,
                           label=f"Temperature_{channel}",
                           unit='K',
                           vals=Numbers(0, 300),
                           get_cmd=None, set_cmd=None)

        self.add_parameter(name='dummy_multi_parameter',
                           parameter_class=MultiSetPointParam)

        self.add_parameter(name='dummy_scalar_multi_parameter',
                           parameter_class=MultiScalarParam)

        self.add_parameter(name='dummy_2d_multi_parameter',
                           parameter_class=Multi2DSetPointParam)

        self.add_parameter(name='dummy_2d_multi_parameter_2',
                           parameter_class=Multi2DSetPointParam2Sizes)

        self.add_parameter(name='dummy_array_parameter',
                           parameter_class=ArraySetPointParam)

        self.add_parameter(name='dummy_complex_array_parameter',
                           parameter_class=ComplexArraySetPointParam)

        self.add_parameter('dummy_start',
                           initial_value=0,
                           unit='some unit',
                           label='f start',
                           vals=Numbers(0, 1e3),
                           get_cmd=None,
                           set_cmd=None)

        self.add_parameter('dummy_stop',
                           unit='some unit',
                           label='f stop',
                           vals=Numbers(1, 1e3),
                           get_cmd=None,
                           set_cmd=None)

        self.add_parameter('dummy_n_points',
                           unit='',
                           vals=Numbers(1, 1e3),
                           get_cmd=None,
                           set_cmd=None)

        self.add_parameter('dummy_start_2',
                           initial_value=0,
                           unit='some unit',
                           label='f start',
                           vals=Numbers(0, 1e3),
                           get_cmd=None,
                           set_cmd=None)

        self.add_parameter('dummy_stop_2',
                           unit='some unit',
                           label='f stop',
                           vals=Numbers(1, 1e3),
                           get_cmd=None,
                           set_cmd=None)

        self.add_parameter('dummy_n_points_2',
                           unit='',
                           vals=Numbers(1, 1e3),
                           get_cmd=None,
                           set_cmd=None)

        self.add_parameter('dummy_sp_axis',
                           unit='some unit',
                           label='Dummy sp axis',
                           parameter_class=GeneratedSetPoints,
                           startparam=self.dummy_start,
                           stopparam=self.dummy_stop,
                           numpointsparam=self.dummy_n_points,
                           vals=Arrays(shape=(self.dummy_n_points,)))

        self.add_parameter('dummy_sp_axis_2',
                           unit='some unit',
                           label='Dummy sp axis',
                           parameter_class=GeneratedSetPoints,
                           startparam=self.dummy_start_2,
                           stopparam=self.dummy_stop_2,
                           numpointsparam=self.dummy_n_points_2,
                           vals=Arrays(shape=(self.dummy_n_points_2,)))

        self.add_parameter(name='dummy_parameter_with_setpoints',
                           label='Dummy Parameter with Setpoints',
                           unit='some other unit',
                           setpoints=(self.dummy_sp_axis,),
                           vals=Arrays(shape=(self.dummy_n_points,)),
                           parameter_class=DummyParameterWithSetpoints1D)

        self.add_parameter(name='dummy_parameter_with_setpoints_2d',
                           label='Dummy Parameter with Setpoints',
                           unit='some other unit',
                           setpoints=(self.dummy_sp_axis,self.dummy_sp_axis_2),
                           vals=Arrays(shape=(self.dummy_n_points,self.dummy_n_points_2)),
                           parameter_class=DummyParameterWithSetpoints2D)

        self.add_parameter(name='dummy_text',
                           label='Dummy text',
                           unit='text unit',
                           initial_value='thisisastring',
                           set_cmd=None,
                           vals=Strings())

        self.add_parameter(name='dummy_complex',
                           label='Dummy complex',
                           unit='complex unit',
                           initial_value=1+1j,
                           set_cmd=None,
                           vals=ComplexNumbers())

        self.add_parameter(name='dummy_parameter_with_setpoints_complex',
                           label='Dummy Parameter with Setpoints complex',
                           unit='some other unit',
                           setpoints=(self.dummy_sp_axis,),
                           vals=Arrays(shape=(self.dummy_n_points,),
                                       valid_types=(np.complexfloating,)),
                           parameter_class=DummyParameterWithSetpointsComplex)

        self.add_function(name='log_my_name',
                          call_cmd=partial(log.debug, f'{name}'))


class DummyChannelInstrument(Instrument):
    """
    Dummy instrument with channels
    """

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

        channels = ChannelList(self, "TempSensors", DummyChannel, snapshotable=False)
        for chan_name in ('A', 'B', 'C', 'D', 'E', 'F'):
            channel = DummyChannel(self, f'Chan{chan_name}', chan_name)
            channels.append(channel)
            self.add_submodule(chan_name, channel)
        self.add_submodule("channels", channels)


class MultiGetter(MultiParameter):
    """
    Test parameters with complicated return values
    instantiate with kwargs::

        MultiGetter(name1=return_val1, name2=return_val2)

    to set the names and (constant) return values of the
    pieces returned. Each return_val can be any array-like
    object
    eg::

        MultiGetter(one=1, onetwo=(1, 2))

    """
    def __init__(self, **kwargs):
        names = tuple(sorted(kwargs.keys()))
        self._return = tuple(kwargs[k] for k in names)
        shapes = tuple(np.shape(v) for v in self._return)
        super().__init__(name='multigetter', names=names, shapes=shapes)

    def get_raw(self):
        return self._return


class MultiSetPointParam(MultiParameter):
    """
    Multiparameter which only purpose it to test that units, setpoints
    and so on are copied correctly to the individual arrays in the datarray.
    """

    def __init__(self, instrument=None, name="multi_setpoint_param", **kwargs):
        shapes = ((5,), (5,))
        names = ('multi_setpoint_param_this', 'multi_setpoint_param_that')
        labels = ('this label', 'that label')
        units = ('this unit', 'that unit')
        sp_base = tuple(np.linspace(5, 9, 5))
        setpoints = ((sp_base,), (sp_base,))
        setpoint_names = (
            ("multi_setpoint_param_this_setpoint",),
            ("multi_setpoint_param_this_setpoint",),
        )
        setpoint_labels = (("this setpoint",), ("this setpoint",))
        setpoint_units = (("this setpointunit",), ("this setpointunit",))
        super().__init__(
            name,
            names,
            shapes,
            instrument=instrument,
            labels=labels,
            units=units,
            setpoints=setpoints,
            setpoint_labels=setpoint_labels,
            setpoint_names=setpoint_names,
            setpoint_units=setpoint_units,
            **kwargs,
        )

    def get_raw(self):
        items = (np.zeros(5), np.ones(5))
        return items


class Multi2DSetPointParam(MultiParameter):
    """
    Multiparameter which only purpose it to test that units, setpoints
    and so on are copied correctly to the individual arrays in the datarray.
    """

    def __init__(self, instrument=None, name="multi_2d_setpoint_param", **kwargs):
        shapes = ((5, 3), (5, 3))
        names = ('this', 'that')
        labels = ('this label', 'that label')
        units = ('this unit', 'that unit')
        sp_base_1 = tuple(np.linspace(5, 9, 5))
        sp_base_2 = tuple(np.linspace(9, 11, 3))
        array_setpoints = setpoint_generator(sp_base_1, sp_base_2)
        setpoints = (array_setpoints, array_setpoints)
        setpoint_names = (
            (
                "multi_2d_setpoint_param_this_setpoint",
                "multi_2d_setpoint_param_that_setpoint",
            ),
            (
                "multi_2d_setpoint_param_this_setpoint",
                "multi_2d_setpoint_param_that_setpoint",
            ),
        )
        setpoint_labels = (
            ("this setpoint", "that setpoint"),
            ("this setpoint", "that setpoint"),
        )
        setpoint_units = (
            ("this setpointunit", "that setpointunit"),
            ("this setpointunit", "that setpointunit"),
        )
        super().__init__(
            name,
            names,
            shapes,
            instrument=instrument,
            labels=labels,
            units=units,
            setpoints=setpoints,
            setpoint_labels=setpoint_labels,
            setpoint_names=setpoint_names,
            setpoint_units=setpoint_units,
            **kwargs,
        )

    def get_raw(self):
        items = (np.zeros((5, 3)), np.ones((5, 3)))
        return items



class Multi2DSetPointParam2Sizes(MultiParameter):
    """
    Multiparameter for testing containing individual parameters with different
    shapes.
    """

    def __init__(self, instrument=None, name="multi_2d_setpoint_param", **kwargs):
        shapes = ((5, 3), (2, 7))
        names = ('this_5_3', 'this_2_7')
        labels = ('this label', 'that label')
        units = ('this unit', 'that unit')
        sp_base_1_1 = tuple(np.linspace(5, 9, 5))
        sp_base_2_1 = tuple(np.linspace(9, 11, 3))
        array_setpoints_1 = setpoint_generator(sp_base_1_1, sp_base_2_1)
        sp_base_1_2 = tuple(np.linspace(5, 9, 2))
        sp_base_2_2 = tuple(np.linspace(9, 11, 7))
        array_setpoints_2 = setpoint_generator(sp_base_1_2, sp_base_2_2)
        setpoints = (array_setpoints_1, array_setpoints_2)
        setpoint_names = (
            (
                "multi_2d_setpoint_param_this_setpoint_1",
                "multi_2d_setpoint_param_that_setpoint_1",
            ),
            (
                "multi_2d_setpoint_param_this_setpoint_2",
                "multi_2d_setpoint_param_that_setpoint_2",
            ),
        )
        setpoint_labels = (
            ("this setpoint 1", "that setpoint 1"),
            ("this setpoint 2", "that setpoint 2"),
        )
        setpoint_units = (
            ("this setpointunit", "that setpointunit"),
            ("this setpointunit", "that setpointunit"),
        )
        super().__init__(
            name,
            names,
            shapes,
            instrument=instrument,
            labels=labels,
            units=units,
            setpoints=setpoints,
            setpoint_labels=setpoint_labels,
            setpoint_names=setpoint_names,
            setpoint_units=setpoint_units,
            **kwargs,
        )

    def get_raw(self):
        items = (np.zeros((5, 3)), np.ones((2, 7)))
        return items


class MultiScalarParam(MultiParameter):
    """
    Multiparameter whos elements are scalars i.e. similar to
    Parameter with no setpoints etc.
    """

    def __init__(self, instrument=None, name="multiscalarparameter", **kwargs):
        shapes = ((), ())
        names = ('thisparam', 'thatparam')
        labels = ('thisparam label', 'thatparam label')
        units = ('thisparam unit', 'thatparam unit')
        setpoints = ((), ())
        super().__init__(
            name,
            names,
            shapes,
            instrument=instrument,
            labels=labels,
            units=units,
            setpoints=setpoints,
            **kwargs,
        )

    def get_raw(self):
        items = (0, 1)
        return items


class ArraySetPointParam(ArrayParameter):
    """
    Arrayparameter which only purpose it to test that units, setpoints
    and so on are copied correctly to the individual arrays in the datarray.
    """

    def __init__(self, instrument=None, name="array_setpoint_param", **kwargs):
        shape = (5,)
        label = 'this label'
        unit = 'this unit'
        sp_base = tuple(np.linspace(5, 9, 5))
        setpoints = (sp_base,)
        setpoint_names = ("array_setpoint_param_this_setpoint",)
        setpoint_labels = ("this setpoint",)
        setpoint_units = ("this setpointunit",)
        super().__init__(
            name,
            shape,
            instrument,
            label=label,
            unit=unit,
            setpoints=setpoints,
            setpoint_labels=setpoint_labels,
            setpoint_names=setpoint_names,
            setpoint_units=setpoint_units,
            **kwargs,
        )

    def get_raw(self):
        item = np.ones(5) + 1
        return item


class ComplexArraySetPointParam(ArrayParameter):
    """
    Arrayparameter that returns complex numbers
    """

    def __init__(self, instrument=None, name="testparameter", **kwargs):
        shape = (5,)
        label = 'this label'
        unit = 'this unit'
        sp_base = tuple(np.linspace(5, 9, 5))
        setpoints = (sp_base,)
        setpoint_names = ("this_setpoint",)
        setpoint_labels = ("this setpoint",)
        setpoint_units = ("this setpointunit",)
        super().__init__(
            name,
            shape,
            instrument,
            label=label,
            unit=unit,
            setpoints=setpoints,
            setpoint_labels=setpoint_labels,
            setpoint_names=setpoint_names,
            setpoint_units=setpoint_units,
            **kwargs,
        )

    def get_raw(self):
        item = np.arange(5) - 1j*np.arange(5)
        return item


class GeneratedSetPoints(Parameter):
    """
    A parameter that generates a setpoint array from start, stop and num points
    parameters.
    """
    def __init__(self, startparam, stopparam, numpointsparam, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._startparam = startparam
        self._stopparam = stopparam
        self._numpointsparam = numpointsparam

    def get_raw(self):
        return np.linspace(self._startparam(), self._stopparam(),
                              self._numpointsparam())


class DummyParameterWithSetpoints1D(ParameterWithSetpoints):
    """
    Dummy parameter that returns data with a shape based on the
    `dummy_n_points` parameter in the instrument.
    """

    def get_raw(self):
        npoints = self.instrument.dummy_n_points()
        return np.random.rand(npoints)


class DummyParameterWithSetpoints2D(ParameterWithSetpoints):
    """
    Dummy parameter that returns data with a shape based on the
    `dummy_n_points` and `dummy_n_points_2` parameters in the instrument.
    """

    def get_raw(self):
        npoints = self.instrument.dummy_n_points()
        npoints_2 = self.instrument.dummy_n_points_2()
        return np.random.rand(npoints, npoints_2)



class DummyParameterWithSetpointsComplex(ParameterWithSetpoints):
    """
    Dummy parameter that returns data with a shape based on the
    `dummy_n_points` parameter in the instrument. Returns Complex values
    """

    def get_raw(self):
        npoints = self.instrument.dummy_n_points()
        return np.random.rand(npoints) + 1j*np.random.rand(npoints)


def setpoint_generator(*sp_bases):
    """
    Helper function to generate setpoints in the format that ArrayParameter
    (and MultiParameter) expects

    Args:
        *sp_bases:

    Returns:

    """
    setpoints = []
    for i, sp_base in enumerate(sp_bases):
        if i == 0:
            setpoints.append(sp_base)
        else:
            repeats = [len(sp) for sp in sp_bases[:i]]
            repeats.append(1)
            setpoints.append(np.tile(sp_base, repeats))

    return tuple(setpoints)


class SnapShotTestInstrument(Instrument):
    """
    A highly specialized dummy instrument for testing the snapshot. Used by
    test_snapshot.py

    Args:
        name: name for the instrument
        params: parameter names. The instrument will have these as parameters
        params_to_skip: parameters to skip updating in the snapshot. Must be
            a subset of params
    """

    def __init__(self, name: str, params: Sequence[str] = ('v1', 'v2', 'v3'),
                 params_to_skip: Sequence[str] = ('v2')):

        super().__init__(name)

        if not(set(params_to_skip).issubset(params)):
            raise ValueError('Invalid input; params_to_skip must be a subset '
                             'of params')

        self._params_to_skip = params_to_skip
        self._params = params

        # dict to keep track of how many time 'get' has been called on each
        # parameter. Useful for testing params_to_skip_update in the snapshot
        self._get_calls = {p: 0 for p in params}

        for p_name in params:

            self.add_parameter(p_name, label=f'{name} Label', unit='V',
                               set_cmd=None,
                               get_cmd=partial(self._getter, p_name))

    def _getter(self, name: str):
        val = self.parameters[name].cache.get(get_if_invalid=False)
        self._get_calls[name] += 1
        return val

    def snapshot_base(self, update: Optional[bool] = True,
                      params_to_skip_update: Optional[Sequence[str]] = None
                      ) -> Dict[Any, Any]:
        if params_to_skip_update is None:
            params_to_skip_update = self._params_to_skip
        snap = super().snapshot_base(
            update=update, params_to_skip_update=params_to_skip_update)
        return snap


class MockField(Instrument):

    def __init__(
            self,
            name: str,
            vals: Numbers = Numbers(min_value=-1., max_value=1.),
            **kwargs):
        """Mock instrument for emulating a magnetic field axis

        Args:
            name (str): Instrument name
            vals (Numbers, optional): Soft limits. Defaults to Numbers(min_value=-1., max_value=1.).
        """
        super().__init__(name=name, **kwargs)
        self._field = 0.0
        self.add_parameter("field",
                           parameter_class=Parameter,
                           initial_value=0.0,
                           unit='T',
                           vals=vals,
                           get_cmd=self.get_field, set_cmd=self.set_field)
        self.add_parameter("ramp_rate",
                           parameter_class=Parameter,
                           initial_value=0.1,
                           unit='T/min',
                           get_cmd=None, set_cmd=None)
        self._ramp_start_time: Optional[float] = None
        self._wait_time: Optional[float] = None
        self._fr = self._field_ramp()
        next(self._fr)

    def get_field(self):
        """
        This method is automatically wrapped to
        provide a ``get`` method on the parameter instance.
        """
        if self._ramp_start_time:
            _time_since_start = time.time() - self._ramp_start_time
            val = self._fr.send(_time_since_start)
            next(self._fr)
            self._field = val
        return self._field

    def set_field(self, value, block: bool = True):
        if self._field == value:
            return value

        wait_time = 60. * np.abs(self._field - value) / self.ramp_rate()
        self._wait_time = wait_time
        self._sign = np.sign(value - self._field)
        self._start_field = self._field
        self._target_field = value
        self._ramp_start_time = time.time()

        if block:
            time.sleep(wait_time)
            self._field = value
            return value

    def _field_ramp_fcn(self, _time: float):
        if self._wait_time is None:
            return self._field
        elif _time <= 0.0:
            return self._start_field
        elif _time >= self._wait_time:
            return self._target_field
        dfield = self.ramp_rate() * _time / 60.0
        return self._start_field + self._sign * dfield

    def _field_ramp(self):
        """
        Yields field for a given point in time
        """
        while True:
            _time = yield
            if _time is None:
                _time = 0.0

            yield float(self._field_ramp_fcn(_time))


class MockLockin(Instrument):

    def __init__(
            self,
            name: str,
            **kwargs):
        super().__init__(name=name, **kwargs)
        self.add_parameter("X",
                           parameter_class=Parameter,
                           initial_value=1e-3,
                           unit='V',
                           get_cmd=None, set_cmd=None)
        self.add_parameter("Y",
                           parameter_class=Parameter,
                           initial_value=1e-5,
                           unit='V',
                           get_cmd=None, set_cmd=None)
        self.add_parameter("frequency",
                           parameter_class=Parameter,
                           initial_value=125.,
                           unit='Hz',
                           get_cmd=None, set_cmd=None)
        self.add_parameter("amplitude",
                           parameter_class=Parameter,
                           initial_value=0.,
                           unit='V',
                           get_cmd=None, set_cmd=None)
        self.add_parameter("phase",
                           parameter_class=Parameter,
                           initial_value=0.,
                           unit='deg',
                           get_cmd=None, set_cmd=None)
        self.add_parameter("time_constant",
                           parameter_class=Parameter,
                           initial_value=1.e-3,
                           unit='s',
                           get_cmd=None, set_cmd=None)


class MockDACChannel(InstrumentChannel):
    """
    A single dummy channel implementation
    """

    def __init__(self, parent, name, num):
        super().__init__(parent, name)

        self._num = num
        self.add_parameter('voltage',
                           parameter_class=Parameter,
                           initial_value=0.,
                           label=f"Voltage_{name}",
                           unit='V',
                           vals=Numbers(-2., 2.),
                           get_cmd=None, set_cmd=None)
        self.add_parameter('dac_output',
                           parameter_class=Parameter,
                           initial_value="off",
                           vals=OnOff(),
                           get_cmd=None, set_cmd=None)
        self.add_parameter('smc',
                           parameter_class=Parameter,
                           initial_value="off",
                           vals=OnOff(),
                           get_cmd=None, set_cmd=None)
        self.add_parameter('bus',
                           parameter_class=Parameter,
                           initial_value="off",
                           vals=OnOff(),
                           get_cmd=None, set_cmd=None)
        self.add_parameter('gnd',
                           parameter_class=Parameter,
                           initial_value="off",
                           vals=OnOff(),
                           get_cmd=None, set_cmd=None)

    def channel_number(self):
        return self._num


class MockDAC(Instrument):

    def __init__(
        self,
        name: str = 'mdac',
        num_channels: int = 10,
        **kwargs):

        """
        Create a dummy instrument that can be used for testing

        Args:
            name: name for the instrument
            gates: list of names that is used to create parameters for
                            the instrument
        """
        super().__init__(name, **kwargs)

        # make gates
        channels = ChannelList(self, "channels", MockDACChannel)
        for n in range(num_channels):
            num = str(n + 1).zfill(2)
            chan_name = f"ch{num}"
            channel = MockDACChannel(parent=self, name=chan_name, num=num)
            channels.append(channel)
            self.add_submodule(chan_name, channel)
        self.add_submodule("channels", channels)


class MockCustomChannel(InstrumentChannel):
    def __init__(
        self,
        parent: InstrumentBase,
        name: str,
        channel: Union[str, InstrumentChannel],
        current_valid_range: Optional[List[float]] = None,
    ) -> None:
        """
        A custom instrument channel emulating an existing channel.

        It adds a parameter not found in the original channel, the
        current_valid_range.
        Args:
            parent: Instrument to which the original channel belongs to,
                usually a dac.
            name: Name of channel.
            channel: The original instrument channel.
            current_valid_range: Voltage range the channel is expected to show
                interesting features. It's just an example of an additional
                parameter a regular instrument channel does not have.
        """
        if isinstance(channel, str):
            _, channel_name = channel.split(".")
            instr_channel = getattr(parent, channel_name)
            self._dac_channel = instr_channel
        elif isinstance(channel, InstrumentChannel):
            self._dac_channel = channel
        else:
            raise ValueError('Unknown input type for "channel".')

        super().__init__(parent, name)

        if current_valid_range is None:
            current_valid_range = []
        super().add_parameter(
            name="current_valid_range",
            label=f"{name} valid voltage range",
            initial_value=current_valid_range,
            vals=ValidatorSequence(Numbers(), length=2),
            get_cmd=None,
            set_cmd=None,
        )

        self.add_parameter(
            "voltage",
            parameter_class=Parameter,
            initial_value=0.0,
            label=f"Voltage_{name}",
            unit="V",
            vals=Numbers(-2.0, 2.0),
            get_cmd=None,
            set_cmd=None,
        )



"""
This file holds plotting utility functions that are
independent of the dataset from which to be plotted.
For the current dataset see :class:`qcodes.dataset.plotting`
For the legacy dataset see :class:`qcodes.plots`
"""
import copy
import logging
from typing import Tuple, Union, Optional, Any, cast, Set, Dict
from collections import OrderedDict
import numpy as np
import matplotlib
import matplotlib.colorbar
import matplotlib.collections
import qcodes

log = logging.getLogger(__name__)


"""
General functions (independent of plotting framework)
"""

# turn off limiting percentiles by default
DEFAULT_PERCENTILE = (50, 50)


def auto_range_iqr(data_array: np.ndarray,
                   cutoff_percentile: Union[
                       Tuple[float, float], float] = DEFAULT_PERCENTILE
                   ) -> Tuple[float, float]:
    """
    Get the min and max range of the provided array that excludes outliers
    following the IQR rule.

    This function computes the inter-quartile-range (IQR), defined by Q3-Q1,
    i.e. the percentiles for 75% and 25% of the distribution. The region
    without outliers is defined by [Q1-1.5*IQR, Q3+1.5*IQR].

    Args:
        data_array: Numpy array of arbitrary dimension containing the
            statistical data.
        cutoff_percentile: Percentile of data that may maximally be clipped
            on both sides of the distribution.
            If given a tuple (a,b) the percentile limits will be a and 100-b.

    Returns:
        region limits [vmin, vmax]
    """
    if isinstance(cutoff_percentile, tuple):
        t = cutoff_percentile[0]
        b = cutoff_percentile[1]
    else:
        t = cutoff_percentile
        b = cutoff_percentile
    z = data_array.flatten()
    zmax = np.nanmax(z)
    zmin = np.nanmin(z)
    zrange = zmax-zmin
    pmin, q3, q1, pmax = np.nanpercentile(z, [b, 75, 25, 100-t])
    IQR = q3-q1
    # handle corner case of all data zero, such that IQR is zero
    # to counter numerical artifacts do not test IQR == 0, but IQR on its
    # natural scale (zrange) to be smaller than some very small number.
    # also test for zrange to be 0.0 to avoid division by 0.
    # all This is possibly to careful...
    if zrange == 0.0 or IQR/zrange < 1e-8:
        vmin = zmin
        vmax = zmax
    else:
        vmin = max(q1 - 1.5*IQR, zmin)
        vmax = min(q3 + 1.5*IQR, zmax)
        # do not clip more than cutoff_percentile:
        vmin = min(vmin, pmin)
        vmax = max(vmax, pmax)
    return vmin, vmax


# Matplotlib functions

DEFAULT_COLOR_OVER = 'Magenta'
DEFAULT_COLOR_UNDER = 'Cyan'


def _set_colorbar_extend(colorbar: matplotlib.colorbar.Colorbar,
                         extend: str) -> None:
    """
    Workaround for a missing setter for the extend property of a matplotlib
    colorbar.

    The colorbar object in matplotlib has no setter method and setting the
    colorbar extend does not take any effect.
    Calling a subsequent update will cause a runtime
    error because of the internal implementation of the rendering of the
    colorbar. To circumvent this we need to manually specify the property
    `_inside`, which is a slice that describes which of the colorbar levels
    lie inside of the box and it is thereby dependent on the extend.

    Args:
        colorbar: the colorbar for which to set the extend
        extend: the desired extend ('neither', 'both', 'min' or 'max')
    """
    colorbar.extend = extend
    _slice_dict = {'neither': slice(0, None),
                   'both': slice(1, -1),
                   'min': slice(1, None),
                   'max': slice(0, -1)}
    colorbar._inside = _slice_dict[extend]


def apply_color_scale_limits(colorbar: matplotlib.colorbar.Colorbar,
                             new_lim: Tuple[Optional[float], Optional[float]],
                             data_lim: Optional[Tuple[float, float]] = None,
                             data_array: Optional[np.ndarray] = None,
                             color_over: Optional[Any] = DEFAULT_COLOR_OVER,
                             color_under: Optional[Any] = DEFAULT_COLOR_UNDER
                             ) -> None:
    """
    Applies limits to colorscale and updates extend.

    This function applies the limits `new_lim` to the heatmap plot associated
    with the provided `colorbar`, updates the colorbar limits, and also adds
    the colorbar clipping indicators in form of small triangles on the top and
    bottom of the colorbar, according to where the limits are exceeded.

    Args:
        colorbar: The actual colorbar to be updated.
        new_lim: 2-tuple of the desired minimum and maximum value of the color
            scale. If any is `None` it will be left unchanged.
        data_lim: 2-tuple of the actual minimum and maximum value of the data.
            If left out the minimum and maximum are deduced from the provided
            data, or the data associated with the colorbar.
        data_array: Numpy array containing the data to be considered for
            scaling. Must be left out if `data_lim` is provided. If neither is
            provided the data associated with the colorbar is used.
        color_over: Matplotlib color representing the datapoints clipped by the
            upper limit.
        color_under: Matplotlib color representing the datapoints clipped by
            the lower limit.

    Raise:
        RuntimeError: If not received mesh data. Or if you specified both
        `data_lim` and `data_array`.
    """
    # browse the input data and make sure that `data_lim` and `new_lim` are
    # available
    if not isinstance(colorbar.mappable, matplotlib.collections.QuadMesh):
        raise RuntimeError('Can only scale mesh data, but received '
                           f'"{type(colorbar.mappable)}" instead')
    if data_lim is None:
        if data_array is None:
            data_array = colorbar.mappable.get_array()
        data_lim = np.nanmin(data_array), np.nanmax(data_array)
    else:
        if data_array is not None:
            raise RuntimeError('You may not specify `data_lim` and `data_array` '
                               'at the same time. Please refer to the docstring of '
                               '`apply_color_scale_limits for details:\n\n`'
                               + str(apply_color_scale_limits.__doc__))
        else:
            data_lim = cast(Tuple[float, float], tuple(sorted(data_lim)))
    # if `None` is provided in the new limits don't change this limit
    vlim = [new or old for new, old in zip(new_lim, colorbar.mappable.get_clim())]
    # sort limits in case they were given in a wrong order
    vlim = sorted(vlim)
    # detect exceeding colorscale and apply new limits
    exceeds_min, exceeds_max = (data_lim[0] < vlim[0],
                                data_lim[1] > vlim[1])
    if exceeds_min and exceeds_max:
        extend = 'both'
    elif exceeds_min:
        extend = 'min'
    elif exceeds_max:
        extend = 'max'
    else:
        extend = 'neither'
    _set_colorbar_extend(colorbar, extend)
    cmap = copy.copy(colorbar.mappable.get_cmap())
    cmap.set_over(color_over)
    cmap.set_under(color_under)
    colorbar.mappable.set_cmap(cmap)
    colorbar.mappable.set_clim(vlim)


def apply_auto_color_scale(colorbar: matplotlib.colorbar.Colorbar,
                           data_array: Optional[np.ndarray] = None,
                           cutoff_percentile: Union[Tuple[
                               float, float], float] = DEFAULT_PERCENTILE,
                           color_over: Optional[Any] = DEFAULT_COLOR_OVER,
                           color_under: Optional[Any] = DEFAULT_COLOR_UNDER
                           ) -> None:
    """
    Sets the color limits such that outliers are disregarded.

    This method combines the automatic color scaling from
    :meth:`auto_range_iqr` with the color bar setting from
    :meth:`apply_color_scale_limits`.
    If you want to adjust the color scale based on the configuration file
    `qcodesrc.json`, use :meth:`auto_color_scale_from_config`, which is used
    In :func:`qcodes.dataset.plotting.plot_by_id`.

    Args:
        colorbar: The matplotlib colorbar to which to apply.
        data_array: The data on which the statistical analysis is based. If
            left out, the data associated with the `colorbar` is used
        cutoff_percentile: Percentile of data that may maximally be clipped
            on both sides of the distribution.
            If given a tuple (a,b) the percentile limits will be a and 100-b.
        color_over: Matplotlib color representing the datapoints clipped by the
            upper limit.
        color_under: Matplotlib color representing the datapoints clipped by
            the lower limit.

    Raises:
        RuntimeError: If not mesh data.
    """
    if data_array is None:
        if not isinstance(colorbar.mappable, matplotlib.collections.QuadMesh):
            raise RuntimeError('Can only scale mesh data.')
        data_array = colorbar.mappable.get_array()
    new_lim = auto_range_iqr(data_array, cutoff_percentile)
    apply_color_scale_limits(colorbar, new_lim=new_lim, data_array=data_array,
                             color_over=color_over, color_under=color_under)


def auto_color_scale_from_config(colorbar: matplotlib.colorbar.Colorbar,
                                 auto_color_scale: Optional[bool] = None,
                                 data_array: Optional[np.ndarray] = None,
                                 cutoff_percentile: Optional[Union[Tuple[
                                     float, float], float]] = DEFAULT_PERCENTILE,
                                 color_over: Optional[Any] = None,
                                 color_under: Optional[Any] = None,
                                 ) -> None:
    """
    Sets the color limits such that outliers are disregarded, depending on
    the configuration file `qcodesrc.json`. If optional arguments are
    passed the config values are overridden.

    Args:
        colorbar: The colorbar to scale.
        auto_color_scale: Enable smart colorscale. If `False` nothing happens.
            Default value is read from
            ``config.plotting.auto_color_scale.enabled``.
        data_array: Numpy array containing the data to be considered for
            scaling.
        cutoff_percentile: The maxiumum percentile that is cut from the data.
            Default value is read from
            ``config.plotting.auto_color_scale.cutoff_percentile``.
        color_over: Matplotlib color representing the datapoints clipped
            by the upper limit. Default value is read from
            ``config.plotting.auto_color_scale.color_over``.
        color_under: Matplotlib color representing the datapoints clipped
            by the lower limit. Default value is read from
            ``config.plotting.auto_color_scale.color_under``.
    """
    if colorbar is None:
        log.warning('"auto_color_scale_from_config" did not receive a colorbar '
                    'for scaling. Are you trying to scale a plot without '
                    'colorbar?')
        return
    if auto_color_scale is None:
        auto_color_scale = qcodes.config.plotting.auto_color_scale.enabled
    if not auto_color_scale:
        return
    if color_over is None:
        color_over = qcodes.config.plotting.auto_color_scale.color_over
    if color_under is None:
        color_under = qcodes.config.plotting.auto_color_scale.color_under
    if cutoff_percentile is None:
        cutoff_percentile = cast(
            Tuple[float, float],
            tuple(qcodes.config.plotting.auto_color_scale.cutoff_percentile))

    apply_auto_color_scale(colorbar, data_array, cutoff_percentile,
                           color_over, color_under)


_UNITS_FOR_RESCALING: Set[str] = {
    # SI units (without some irrelevant ones like candela)
    # 'kg' is not included because it is 'kilo' and rarely used
    'm', 's', 'A', 'K', 'mol', 'rad', 'Hz', 'N', 'Pa', 'J',
    'W', 'C', 'V', 'F', 'ohm', 'Ohm', 'Ω',
    '\N{GREEK CAPITAL LETTER OMEGA}', 'S', 'Wb', 'T', 'H',
    # non-SI units as well, for convenience
    'eV', 'g'
}

_ENGINEERING_PREFIXES: Dict[int, str] = OrderedDict({
    -24: "y",
    -21: "z",
    -18: "a",
    -15: "f",
    -12: "p",
     -9: "n",
     -6: "\N{GREEK SMALL LETTER MU}",
     -3: "m",
      0: "",
      3: "k",
      6: "M",
      9: "G",
     12: "T",
     15: "P",
     18: "E",
     21: "Z",
     24: "Y"
})

_THRESHOLDS: Dict[float, int] = OrderedDict(
    {10**(scale + 3): scale for scale in _ENGINEERING_PREFIXES.keys()})


def find_scale_and_prefix(data: np.ndarray, unit: str) -> Tuple[str, int]:
    """
    Given a numpy array of data and a unit find the best engineering prefix
    and matching scale that best describes the data.

    The units for which unit prefixes are added can be found in
    ``_UNITS_FOR_RESCALING``. For all other units
    the prefix is an exponential scaling factor e.g. ``10^3``.
    If no unit is given (i.e. an empty string) no scaling is performed.

    Args:
        data: A numpy array of data.
        unit: The unit the data is measured in.
            Should not contain a prefix already.

    Returns:
        tuple of prefix and exponential scale

    """
    maxval = np.nanmax(np.abs(data))
    if unit in _UNITS_FOR_RESCALING:
        for threshold, scale in _THRESHOLDS.items():
            if maxval == 0.0:
                # handle all zero data
                selected_scale = 0
                prefix = _ENGINEERING_PREFIXES[selected_scale]
                break
            if maxval < threshold:
                selected_scale = scale
                prefix = _ENGINEERING_PREFIXES[scale]
                break
        else:
            # here, maxval is larger than the largest threshold
            largest_scale = max(list(_ENGINEERING_PREFIXES.keys()))
            selected_scale = largest_scale
            prefix = _ENGINEERING_PREFIXES[largest_scale]
    elif unit != "":
        if maxval > 0:
            selected_scale = 3 * (np.floor(np.floor(np.log10(maxval)) / 3))
        else:
            selected_scale = 0
        if selected_scale != 0:
            prefix = f'$10^{{{selected_scale:.0f}}}$ '
        else:
            prefix = ''
    else:
        prefix = ""
        selected_scale = 0
    return prefix, selected_scale



from unittest.mock import MagicMock
import re

import pytest

from qcodes.instrument_drivers.Keysight.keysightb1500.KeysightB1511B import \
    B1511B
from qcodes.instrument_drivers.Keysight.keysightb1500.constants import \
    IOutputRange, IMeasRange

# pylint: disable=redefined-outer-name


@pytest.fixture
def mainframe():
    yield MagicMock()


@pytest.fixture
def smu(mainframe):
    slot_nr = 1
    smu = B1511B(parent=mainframe, name='B1511B', slot_nr=slot_nr)
    yield smu


def test_force_invalid_current_output_range_when_asu_not_present(smu):
    msg = re.escape("Invalid Source Current Output Range")
    with pytest.raises(RuntimeError, match=msg):
        smu.asu_present = True
        smu.asu_present = False
        smu.source_config(IOutputRange.MIN_1pA)


def test_i_measure_range_config_raises_invalid_range_error_when_asu_not_present(
        smu):
    msg = re.escape("8 current measurement range")
    with pytest.raises(RuntimeError, match=msg):
        smu.asu_present = True
        smu.asu_present = False
        smu.i_measure_range_config(IMeasRange.MIN_1pA)



from functools import partial
import logging
from typing import Union, Any

from qcodes import VisaInstrument, validators as vals
from qcodes.instrument.channel import InstrumentChannel
from qcodes.instrument.base import Instrument
from qcodes.instrument_drivers.Keysight.private.error_handling import \
    KeysightErrorQueueMixin

log = logging.getLogger(__name__)


# This is to be the grand unified driver superclass for
# The Keysight/Agilent/HP Waveform generators of series
# 33200, 33500, and 33600


class OutputChannel(InstrumentChannel):
    """
    Class to hold the output channel of a waveform generator
    """
    def __init__(self, parent: Instrument, name: str, channum: int) -> None:
        """
        Args:
            parent: The instrument to which the channel is
                attached.
            name: The name of the channel
            channum: The number of the channel in question (1-2)
        """
        super().__init__(parent, name)

        def val_parser(parser: type, inputstring: str) -> Union[float,int]:
            """
            Parses return values from instrument. Meant to be used when a query
            can return a meaningful finite number or a numeric representation
            of infinity

            Args:
                parser: Either int or float, what to return in finite
                    cases
                inputstring: The raw return value
            """

            inputstring = inputstring.strip()

            if float(inputstring) == 9.9e37:
                output = float('inf')
            else:
                output = float(inputstring)
                if parser == int:
                    output = parser(output)

            return output

        self.model = self._parent.model

        self.add_parameter('function_type',
                           label=f'Channel {channum} function type',
                           set_cmd=f'SOURce{channum}:FUNCtion {{}}',
                           get_cmd=f'SOURce{channum}:FUNCtion?',
                           get_parser=str.rstrip,
                           vals=vals.Enum('SIN', 'SQU', 'TRI', 'RAMP',
                                          'PULS', 'PRBS', 'NOIS', 'ARB',
                                          'DC')
                           )

        self.add_parameter('frequency_mode',
                           label=f'Channel {channum} frequency mode',
                           set_cmd=f'SOURce{channum}:FREQuency:MODE {{}}',
                           get_cmd=f'SOURce{channum}:FREQuency:MODE?',
                           get_parser=str.rstrip,
                           vals=vals.Enum('CW', 'LIST', 'SWEEP', 'FIXED')
                           )

        max_freq = self._parent._max_freqs[self.model]
        self.add_parameter('frequency',
                           label=f'Channel {channum} frequency',
                           set_cmd=f'SOURce{channum}:FREQuency {{}}',
                           get_cmd=f'SOURce{channum}:FREQuency?',
                           get_parser=float,
                           unit='Hz',
                           # TODO: max. freq. actually really tricky
                           vals=vals.Numbers(1e-6, max_freq)
                           )

        self.add_parameter('phase',
                           label=f'Channel {channum} phase',
                           set_cmd=f'SOURce{channum}:PHASe {{}}',
                           get_cmd=f'SOURce{channum}:PHASe?',
                           get_parser=float,
                           unit='deg',
                           vals=vals.Numbers(0, 360)
                           )
        self.add_parameter('amplitude_unit',
                           label=f'Channel {channum} amplitude unit',
                           set_cmd=f'SOURce{channum}:VOLTage:UNIT {{}}',
                           get_cmd=f'SOURce{channum}:VOLTage:UNIT?',
                           vals=vals.Enum('VPP', 'VRMS', 'DBM'),
                           get_parser=str.rstrip
                           )

        self.add_parameter('amplitude',
                           label=f'Channel {channum} amplitude',
                           set_cmd=f'SOURce{channum}:VOLTage {{}}',
                           get_cmd=f'SOURce{channum}:VOLTage?',
                           unit='',  # see amplitude_unit
                           get_parser=float)

        self.add_parameter('offset',
                           label=f'Channel {channum} voltage offset',
                           set_cmd=f'SOURce{channum}:VOLTage:OFFSet {{}}',
                           get_cmd=f'SOURce{channum}:VOLTage:OFFSet?',
                           unit='V',
                           get_parser=float
                           )
        self.add_parameter('output',
                           label=f'Channel {channum} output state',
                           set_cmd=f'OUTPut{channum} {{}}',
                           get_cmd=f'OUTPut{channum}?',
                           val_mapping={'ON': 1, 'OFF': 0}
                           )

        self.add_parameter('ramp_symmetry',
                           label=f'Channel {channum} ramp symmetry',
                           set_cmd=f'SOURce{channum}:FUNCtion:RAMP:SYMMetry {{}}',
                           get_cmd=f'SOURce{channum}:FUNCtion:RAMP:SYMMetry?',
                           get_parser=float,
                           unit='%',
                           vals=vals.Numbers(0, 100)
                           )

        self.add_parameter('pulse_width',
                           label=f"Channel {channum} pulse width",
                           set_cmd=f'SOURce{channum}:FUNCtion:PULSE:WIDTh {{}}',
                           get_cmd=f'SOURce{channum}:FUNCtion:PULSE:WIDTh?',
                           get_parser=float,
                           unit='S')

        # TRIGGER MENU
        self.add_parameter('trigger_source',
                           label=f'Channel {channum} trigger source',
                           set_cmd=f'TRIGger{channum}:SOURce {{}}',
                           get_cmd=f'TRIGger{channum}:SOURce?',
                           vals=vals.Enum('IMM', 'EXT', 'TIM', 'BUS'),
                           get_parser=str.rstrip,
                           )

        self.add_parameter('trigger_slope',
                           label=f'Channel {channum} trigger slope',
                           set_cmd=f'TRIGger{channum}:SLOPe {{}}',
                           get_cmd=f'TRIGger{channum}:SLOPe?',
                           vals=vals.Enum('POS', 'NEG'),
                           get_parser=str.rstrip
                           )

        # Older models do not have all the fancy trigger options
        if self._parent.model[2] in ['5', '6']:
            self.add_parameter('trigger_count',
                               label=f'Channel {channum} trigger count',
                               set_cmd=f'TRIGger{channum}:COUNt {{}}',
                               get_cmd=f'TRIGger{channum}:COUNt?',
                               vals=vals.Ints(1, 1000000),
                               get_parser=partial(val_parser, int)
                               )

            self.add_parameter('trigger_delay',
                               label=f'Channel {channum} trigger delay',
                               set_cmd=f'TRIGger{channum}:DELay {{}}',
                               get_cmd=f'TRIGger{channum}:DELay?',
                               vals=vals.Numbers(0, 1000),
                               get_parser=float,
                               unit='s')

            self.add_parameter('trigger_timer',
                               label=f'Channel {channum} trigger timer',
                               set_cmd=f'TRIGger{channum}:TIMer {{}}',
                               get_cmd=f'TRIGger{channum}:TIMer?',
                               vals=vals.Numbers(1e-6, 8000),
                               get_parser=float)

        # TODO: trigger level doesn't work remotely. Why?

        # output menu
        self.add_parameter('output_polarity',
                           label=f'Channel {channum} output polarity',
                           set_cmd=f'OUTPut{channum}:POLarity {{}}',
                           get_cmd=f'OUTPut{channum}:POLarity?',
                           get_parser=str.rstrip,
                           vals=vals.Enum('NORM', 'INV')
                           )
        # BURST MENU
        self.add_parameter('burst_state',
                           label=f'Channel {channum} burst state',
                           set_cmd=f'SOURce{channum}:BURSt:STATe {{}}',
                           get_cmd=f'SOURce{channum}:BURSt:STATe?',
                           val_mapping={'ON': 1, 'OFF': 0},
                           vals=vals.Enum('ON', 'OFF')
                           )

        self.add_parameter('burst_mode',
                           label=f'Channel {channum} burst mode',
                           set_cmd=f'SOURce{channum}:BURSt:MODE {{}}',
                           get_cmd=f'SOURce{channum}:BURSt:MODE?',
                           get_parser=str.rstrip,
                           val_mapping={'N Cycle': 'TRIG', 'Gated': 'GAT'},
                           vals=vals.Enum('N Cycle', 'Gated')
                           )

        self.add_parameter('burst_ncycles',
                           label=f'Channel {channum} burst no. of cycles',
                           set_cmd=f'SOURce{channum}:BURSt:NCYCles {{}}',
                           get_cmd=f'SOURce{channum}:BURSt:NCYCLes?',
                           get_parser=partial(val_parser, int),
                           vals=vals.MultiType(vals.Ints(1),
                                               vals.Enum('MIN', 'MAX',
                                                         'INF'))
                           )

        self.add_parameter('burst_phase',
                           label=f'Channel {channum} burst start phase',
                           set_cmd=f'SOURce{channum}:BURSt:PHASe {{}}',
                           get_cmd=f'SOURce{channum}:BURSt:PHASe?',
                           vals=vals.Numbers(-360, 360),
                           unit='degrees',
                           get_parser=float
                           )

        self.add_parameter('burst_polarity',
                           label=f'Channel {channum} burst gated polarity',
                           set_cmd=f'SOURce{channum}:BURSt:GATE:POLarity {{}}',
                           get_cmd=f'SOURce{channum}:BURSt:GATE:POLarity?',
                           vals=vals.Enum('NORM', 'INV')
                           )

        self.add_parameter('burst_int_period',
                           label=(f'Channel {channum}' +
                                  ' burst internal period'),
                           set_cmd=f'SOURce{channum}:BURSt:INTernal:PERiod {{}}',
                           get_cmd=f'SOURce{channum}:BURSt:INTernal:PERiod?',
                           unit='s',
                           vals=vals.Numbers(1e-6, 8e3),
                           get_parser=float,
                           docstring=('The burst period is the time '
                                      'between the starts of consecutive '
                                      'bursts when trigger is immediate.')
                           )


class SyncChannel(InstrumentChannel):
    """
    Class to hold the sync output. Has very few parameters for
    single channel instruments
    """

    def __init__(self, parent: Instrument, name: str):

        super().__init__(parent, name)

        self.add_parameter('output',
                           label='Sync output state',
                           set_cmd='OUTPut:SYNC {}',
                           get_cmd='OUTPut:SYNC?',
                           val_mapping={'ON': 1, 'OFF': 0},
                           vals=vals.Enum('ON', 'OFF')
                           )

        if parent.num_channels == 2:

            self.add_parameter('source',
                               label='Source of sync function',
                               set_cmd='OUTPut:SYNC:SOURce {}',
                               get_cmd='OUTPut:SYNC:SOURce?',
                               val_mapping={1: 'CH1', 2: 'CH2'},
                               vals=vals.Enum(1, 2))


class WaveformGenerator_33XXX(KeysightErrorQueueMixin, VisaInstrument):
    """
    QCoDeS driver for the Keysight/Agilent 33XXX series of
    waveform generators
    """

    def __init__(self, name: str, address: str,
                 silent: bool = False, **kwargs: Any):
        """
        Args:
            name: The name of the instrument used internally
                by QCoDeS. Must be unique.
            address: The VISA resource name.
            silent: If True, no connect message is printed.
        """

        super().__init__(name, address, terminator='\n', **kwargs)
        self.model = self.IDN()['model']

        #######################################################################
        # Here go all model specific traits

        # TODO: Fill out this dict with all models
        no_of_channels = {'33210A': 1,
                          '33250A': 1,
                          '33511B': 1,
                          '33512B': 2,
                          '33522B': 2,
                          '33622A': 2,
                          '33510B': 2,
                          }

        self._max_freqs = {'33210A': 10e6,
                           '33511B': 20e6,
                           '33512B': 20e6,
                           '33250A': 80e6,
                           '33522B': 30e6,
                           '33622A': 120e6,
                           '33510B': 20e6,
                          }

        self.num_channels = no_of_channels[self.model]

        for i in range(1, self.num_channels+1):
            channel = OutputChannel(self, f'ch{i}', i)
            self.add_submodule(f'ch{i}', channel)

        sync = SyncChannel(self, 'sync')
        self.add_submodule('sync', sync)

        self.add_function('force_trigger', call_cmd='*TRG')

        self.add_function('sync_channel_phases', call_cmd='PHAS:SYNC')

        if not silent:
            self.connect_message()



from datetime import datetime
from unittest import TestCase

from qcodes.data.location import FormatLocation, SafeFormatter

from .data_mocks import MatchIO


class TestSafeFormatter(TestCase):
    def test_normal_formatting(self):
        formatter = SafeFormatter()
        self.assertEqual(formatter.format('{}{{{}}}', 1, 2), '1{2}')
        self.assertEqual(formatter.format('{apples}{{{oranges}}}',
                                          apples='red', oranges='unicorn'),
                         'red{unicorn}')

    def test_missing(self):
        formatter = SafeFormatter()
        self.assertEqual(formatter.format('{}'), '{0}')
        self.assertEqual(formatter.format('{cheese}', fruit='pears'),
                         '{cheese}')


def _default(time: datetime, formatter: FormatLocation, counter:str, name: str):
    date = time.strftime(formatter.fmt_date)
    mytime = time.strftime(formatter.fmt_time)
    fmted = formatter.formatter.format(formatter.default_fmt,
                                       date=date,
                                       counter=counter,
                                       time=mytime,
                                       name=name)
    return fmted


class TestFormatLocation(TestCase):
    def test_default(self):
        lp = FormatLocation()
        fmt = '%Y-%m-%d/%H-%M-%S'

        name = "name"
        self.assertEqual(lp(MatchIO([]), {'name': name}),
                         _default(datetime.now(), lp, "001", name))

        # counter starts at +1  using MatchIo undocumented magic argument
        start_magic_value = 5
        self.assertEqual(
            lp(MatchIO(["", f"{start_magic_value:03d}"]), {"name": name}),
            _default(datetime.now(), lp, f"{start_magic_value+1:03d}", name),
        )

    def test_fmt_subparts(self):
        lp = FormatLocation(fmt='{date}/{time}', fmt_date='%d-%b-%Y', fmt_time='%I-%M%p',
                            fmt_counter='##{:.1f}~')
        fmt = '%d-%b-%Y/%I-%M%p'

        self.assertEqual(lp(MatchIO([])),
                         datetime.now().strftime(fmt))
        self.assertEqual(lp(MatchIO([]), {'name': 'who?'}),
                         datetime.now().strftime(fmt) + '_who?')

        self.assertEqual(lp(MatchIO([''])),
                         datetime.now().strftime(fmt) + '_##2.0~')
        self.assertEqual(lp(MatchIO(['', '9'])),
                         datetime.now().strftime(fmt) + '_##10.0~')
        self.assertEqual(lp(MatchIO(['', '12345']), {'name': 'you!'}),
                         datetime.now().strftime(fmt) + '_you!_##12346.0~')

    def test_record_call(self):
        lp = FormatLocation(fmt='{date}/#{counter}_{time}_{name}_{label}')

        # counter starts at 1 if it's a regular part of the format string
        expected = datetime.now().strftime('%Y-%m-%d/#001_%H-%M-%S_Joe_fruit')
        self.assertEqual(lp(MatchIO([]), {'name': 'Joe', 'label': 'fruit'}),
                         expected)

        expected = datetime.now().strftime('%Y-%m-%d/#1000_%H-%M-%S_Ga_As')
        self.assertEqual(lp(MatchIO(['999']), {'name': 'Ga', 'label': 'As'}),
                         expected)

        # missing label
        expected = datetime.now().strftime(
            '%Y-%m-%d/#001_%H-%M-%S_Fred_{label}')
        self.assertEqual(lp(MatchIO([]), {'name': 'Fred'}), expected)

    def test_record_override(self):
        # this one needs 'c' filled in at call time
        lp = FormatLocation(fmt='{a}_{b}_{c}', record={'a': 'A', 'b': 'B'})
        io = MatchIO([])

        self.assertEqual(lp(io, {'c': 'C'}), 'A_B_C')
        self.assertEqual(lp(io, {'a': 'aa', 'c': 'cc'}), 'aa_B_cc')
        # extra keys just get discarded
        self.assertEqual(lp(io, {'c': 'K', 'd': 'D'}), 'A_B_K')

        self.assertEqual(lp(MatchIO([])), 'A_B_{c}')
        self.assertEqual(lp(MatchIO([]), {'a': 'AA'}), 'AA_B_{c}')

        # this one has defaults for everything, so nothing is needed at
        # call time, but things can still be overridden
        lp = FormatLocation(fmt='{d}_{e}', record={'d': 'D', 'e': 'E'})

        self.assertEqual(lp(MatchIO([])), 'D_E')
        self.assertEqual(lp(MatchIO([]), {'d': 'T'}), 'T_E')

    def test_errors(self):
        io = MatchIO([])

        # gives the wrong integer (extra 0 at the end!)
        with self.assertRaises(ValueError):
            FormatLocation(fmt_counter='{:03}0')

        # you're not allowed to give a counter. Whatcha trying to do anyway?
        # I'm tempted to say the same about time and date, but will leave them
        # overridable for now.
        with self.assertRaises(KeyError):
            FormatLocation()(io, {'counter': 100})
        with self.assertRaises(KeyError):
            FormatLocation(record={'counter': 100})(io)



import logging
import pytest
import numpy as np

from qcodes.instrument_drivers.tektronix.Keithley_2450 import Keithley2450

import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'Keithley_2450.yaml@sim')


@pytest.fixture(scope='function')
def k2450():
    """
    Create a Keithley 2450 instrument
    """
    driver = Keithley2450('k2450', address='GPIB::2::INSTR', visalib=visalib)
    yield driver
    driver.close()


def test_wrong_mode(caplog):
    """
    Starting an instrument in the wrong mode should result in a warning. Additionally, no
    parameters should be available, other then the parameters "IDN" and "timeout" which
    are created by the Instrument and VisaInstrument parent classes
    """
    with caplog.at_level(logging.WARNING):
        instrument = Keithley2450('wrong_mode', address='GPIB::1::INSTR', visalib=visalib)
        assert "The instrument is in an unsupported language mode." in caplog.text
        assert list(instrument.parameters.keys()) == ["IDN", "timeout"]
        instrument.close()


def test_change_source_function(k2450):
    """
    The parameters available on the source sub-module depend on the function.
    """
    k2450.source.function("current")
    assert k2450.source is k2450.submodules["_source_current"]
    assert hasattr(k2450.source, "current")
    assert not hasattr(k2450.source, "voltage")

    k2450.source.function("voltage")
    assert k2450.source is k2450.submodules["_source_voltage"]
    assert not hasattr(k2450.source, "current")
    assert hasattr(k2450.source, "voltage")

    # to cover all bases :-)
    assert k2450.submodules["_source_current"] is not k2450.submodules["_source_voltage"]


def test_source_change_error(k2450):
    """
    If the sense function is in resistance mode, test that an error is generated
    when we change the source function
    """
    k2450.source.function("voltage")
    k2450.sense.function("resistance")
    with pytest.raises(
        RuntimeError,
        match="Cannot change the source function while sense function is in 'resistance' mode"
    ):
        k2450.source.function("current")

    k2450.sense.function("voltage")
    # after changing the sense function we should be able to change the source function
    k2450.source.function("current")


def test_sense_current_mode(k2450):
    """
    Test that when we are in a sense function, for example, 'current', that
    the sense property is returning to the correct submodule. We also test
    that if we are in a sense function (e.g. 'current') the returned submodule
    does not have inappropriate parameters, (e.g. 'voltage', or 'current')
    """
    sense_functions = {"current", "voltage", "resistance"}
    for sense_function in sense_functions:

        k2450.sense.function(sense_function)
        assert k2450.sense is k2450.submodules[f"_sense_{sense_function}"]
        assert hasattr(k2450.sense, sense_function)

        for other_sense_function in sense_functions.difference({sense_function}):
            assert not hasattr(k2450.sense, other_sense_function)


def test_setpoint_always_follows_source_function(k2450):
    """
    Changing the source and/or sense functions should not confuse the setpoints. These
    should always follow the source module
    """
    n = 100
    sense_modes = np.random.choice(["current", "voltage", "resistance"], n)
    source_modes = np.random.choice(["current", "voltage"], n)

    for sense_mode, source_mode in zip(sense_modes, source_modes):
        k2450.sense.function("voltage")  # In 'resistance' sense mode, we cannot
        # change the source mode by design. Therefore temporarily switch to
        # 'voltage'
        k2450.source.function(source_mode)
        k2450.sense.function(sense_mode)
        assert k2450.sense.sweep.setpoints == (k2450.source.sweep_axis,)


def test_reset_sweep_on_source_change(k2450):
    """
    If we change the source function, we need to run the sweep setup again
    """
    # first set sense to a mode where we are allowed to change source
    k2450.sense.function('current')
    k2450.source.function("voltage")
    k2450.source.sweep_setup(0, 1, 10)
    assert np.alltrue(k2450.source.get_sweep_axis() == np.linspace(0, 1, 10))

    k2450.source.function("current")
    with pytest.raises(ValueError):
        k2450.source.get_sweep_axis()


def test_sweep(k2450):
    """
    Verify that we can start sweeps
    """
    k2450.sense.function("current")
    k2450.source.function("voltage")
    k2450.source.sweep_setup(0, 1, 10)
    k2450.sense.sweep.get()



import pytest

from qcodes.tests.instrument_mocks import DummyInstrument
from qcodes.instrument.parameter import Parameter
import qcodes.utils.validators as vals


class DummyTrackingInstrument(DummyInstrument):

    def __init__(self, name):
        super().__init__(name)
        self.add_parameter("a",
                           set_cmd=None,
                           get_cmd=None)

        # These two parameters mock actual instrument parameters; when first
        # connecting to the instrument, they have the _latest["value"] None.
        # We must call get() on them to get a valid value that we can set
        # them to in the __exit__ method of the context manager
        self.add_parameter("validated_param",
                           initial_cache_value=None,
                           set_cmd=self._vp_setter,
                           get_cmd=self._vp_getter,
                           vals=vals.Enum("foo", "bar"))

        self.add_parameter("parsed_param",
                           initial_cache_value=None,
                           set_cmd=self._pp_setter,
                           get_cmd=self._pp_getter,
                           set_parser=int)

        # A parameter that counts the number of times it has been set
        self.add_parameter("counting_parameter",
                           set_cmd=self._cp_setter,
                           get_cmd=self._cp_getter)

        # the mocked instrument state values of validated_param and
        # parsed_param
        self._vp_value = "foo"
        self._pp_value = 42

        # the counter value for counting_parameter
        self._cp_counter = 0
        self._cp_get_counter = 0

    def _vp_getter(self):
        return self._vp_value

    def _vp_setter(self, value):
        self._vp_value = value

    def _pp_getter(self):
        return self._pp_value

    def _pp_setter(self, value):
        self._pp_value = value

    def _cp_setter(self, value):
        self._cp_counter += 1

    def _cp_getter(self):
        self._cp_get_counter += 1
        return self.counting_parameter.cache._value


@pytest.fixture()
def instrument():
    instrument = DummyTrackingInstrument('dummy_holder')
    try:
        yield instrument
    finally:
        instrument.close()


def test_set_to_none_when_parameter_is_not_captured_yet(instrument):
    counting_parameter = instrument.counting_parameter
    # Pre-conditions:
    assert instrument._cp_counter == 0
    assert instrument._cp_get_counter == 0
    assert counting_parameter.cache._value is None
    assert counting_parameter.get_latest.get_timestamp() is None

    with counting_parameter.set_to(None):
        # The value should not change
        assert counting_parameter.cache._value is None
        # The timestamp of the latest value should not be None anymore
        assert counting_parameter.get_latest.get_timestamp() is not None
        # Set method is not called
        assert instrument._cp_counter == 0
        # Get method is called once
        assert instrument._cp_get_counter == 1

    # The value should not change
    assert counting_parameter.cache._value is None
    # The timestamp of the latest value should still not be None
    assert counting_parameter.get_latest.get_timestamp() is not None
    # Set method is still not called
    assert instrument._cp_counter == 0
    # Get method is still called once
    assert instrument._cp_get_counter == 1


def test_set_to_none_for_not_captured_parameter_but_instrument_has_value():
    # representing instrument here
    instr_value = 'something'
    set_counter = 0

    def set_instr_value(value):
        nonlocal instr_value, set_counter
        instr_value = value
        set_counter += 1

    # make a parameter that is linked to an instrument
    p = Parameter('p', set_cmd=set_instr_value, get_cmd=lambda: instr_value,
                  val_mapping={'foo': 'something', None: 'nothing'})

    # pre-conditions
    assert p.cache._value is None
    assert p.cache._raw_value is None
    assert p.cache.timestamp is None
    assert set_counter == 0

    with p.set_to(None):
        # assertions after entering the context
        assert set_counter == 1
        assert instr_value == 'nothing'
        assert p.cache._value is None
        assert p.cache._raw_value == 'nothing'
        assert p.cache.timestamp is not None

    # assertions after exiting the context
    assert set_counter == 2
    assert instr_value == 'something'
    assert p.cache._value == 'foo'
    assert p.cache._raw_value == 'something'
    assert p.cache.timestamp is not None


def test_none_value(instrument):
    with instrument.a.set_to(3):
        assert instrument.a.get_latest.get_timestamp() is not None
        assert instrument.a.get() == 3
    assert instrument.a.get() is None
    assert instrument.a.get_latest.get_timestamp() is not None


def test_context(instrument):
    instrument.a.set(2)

    with instrument.a.set_to(3):
        assert instrument.a.get() == 3
    assert instrument.a.get() == 2


def test_validated_param(instrument):
    assert instrument.parsed_param.cache._value is None
    assert instrument.validated_param.get_latest() == "foo"
    with instrument.validated_param.set_to("bar"):
        assert instrument.validated_param.get() == "bar"
    assert instrument.validated_param.get_latest() == "foo"
    assert instrument.validated_param.get() == "foo"


def test_parsed_param(instrument):
    assert instrument.parsed_param.cache._value is None
    assert instrument.parsed_param.get_latest() == 42
    with instrument.parsed_param.set_to(1):
        assert instrument.parsed_param.get() == 1
    assert instrument.parsed_param.get_latest() == 42
    assert instrument.parsed_param.get() == 42


def test_number_of_set_calls(instrument):
    """
    Test that with param.set_to(X) does not perform any calls to set if
    the parameter already had the value X
    """
    assert instrument._cp_counter == 0
    instrument.counting_parameter(1)
    assert instrument._cp_counter == 1

    with instrument.counting_parameter.set_to(2):
        pass
    assert instrument._cp_counter == 3

    with instrument.counting_parameter.set_to(1):
        pass
    assert instrument._cp_counter == 3


def test_value_modified_between_context_create_and_enter(instrument):
    p = instrument.a
    p.set(2)
    ctx = p.set_to(5)
    # the parameter value is changed after the context has been created
    # this is the value it should return to after exit.
    p.set(3)
    with ctx:
        assert p() == 5
    assert p() == 3


def test_disallow_changes(instrument):
    instrument.a.set(2)

    with instrument.a.set_to(3, allow_changes=False):
        assert instrument.a() == 3
        assert not instrument.a.settable
        with pytest.raises(TypeError):
            instrument.a.set(5)

    assert instrument.a.settable
    assert instrument.a() == 2


def test_allow_changes(instrument):
    p = instrument.a
    p.set(2)
    with p.set_to(3, allow_changes=True):
        assert p.settable
        assert p() == 3
        p.set(5)
        assert p() == 5

    assert p.settable
    assert p() == 2

    # check that the value gets restored even if entering the context
    # with the current value
    with instrument.a.set_to(2, allow_changes=True):
        assert p() == 2
        p(5)
        assert p() == 5

    assert p.settable
    assert p() == 2


def test_reset_at_exit(instrument):
    p = instrument.a
    p.set(2)
    with p.restore_at_exit():
        p.set(5)
    assert p() == 2


def test_reset_at_exit_with_allow_changes_false(instrument):
    p = instrument.a
    p.set(2)
    with p.restore_at_exit(allow_changes=False):
        with pytest.raises(TypeError):
            p.set(5)
    assert p() == 2



import logging
from functools import partial
from typing import Any, Dict, List, Union, cast

from qcodes import ChannelList, InstrumentChannel, VisaInstrument
from qcodes import validators as vals
from qcodes.utils.helpers import partial_with_docstring

log = logging.getLogger(__name__)


class DG1062Burst(InstrumentChannel):
    """
    Burst commands for the DG1062. We make a separate channel for these to
    group burst commands together.
    """

    def __init__(self, parent: 'DG1062', name: str, channel: int):
        super().__init__(parent, name)
        self.channel = channel

        self.add_parameter(
            "on",
            get_cmd=f":SOUR{channel}:BURS?",
            set_cmd=f":SOUR{channel}:BURS {{}}",
            vals=vals.Enum(0, 1, "ON", "OFF")
        )

        self.add_parameter(
            "polarity",
            get_cmd=f":SOUR{channel}:BURS:GATE:POL?",
            set_cmd=f":SOUR{channel}:BURS:GATE:POL {{}}",
            vals=vals.Enum("NORM", "INV")
        )

        self.add_parameter(
            "period",
            get_cmd=f":SOUR{channel}:BURS:INT:PER?",
            set_cmd=f":SOUR{channel}:BURS:INT:PER {{}}",
            vals=vals.MultiType(
                vals.Numbers(min_value=3E-6, max_value=500),
                vals.Enum("MIN", "MAX")
            )
        )

        self.add_parameter(
            "mode",
            get_cmd=f":SOUR{channel}:BURS:MODE?",
            set_cmd=f":SOUR{channel}:BURS:MODE {{}}",
            vals=vals.Enum("TRIG", "INF", "GAT")
        )

        self.add_parameter(
            "ncycles",
            get_cmd=f":SOUR{channel}:BURS:NCYC?",
            set_cmd=f":SOUR{channel}:BURS:NCYC {{}}",
            vals=vals.Numbers(min_value=1, max_value=500000)
        )

        self.add_parameter(
            "phase",
            get_cmd=f":SOUR{channel}:BURS:PHAS?",
            set_cmd=f":SOUR{channel}:BURS:PHAS {{}}",
            vals=vals.Numbers(min_value=0, max_value=360)
        )

        self.add_parameter(
            "time_delay",
            get_cmd=f":SOUR{channel}:BURS:TDEL?",
            set_cmd=f":SOUR{channel}:BURS:TDEL {{}}",
            vals=vals.Numbers(min_value=0)
        )

        self.add_parameter(
            "trigger_slope",
            get_cmd=f":SOUR{channel}:BURS:TRIG:SLOP?",
            set_cmd=f":SOUR{channel}:BURS:TRIG:SLOP {{}}",
            vals=vals.Enum("POS", "NEG")
        )

        self.add_parameter(
            "source",
            get_cmd=f":SOUR{channel}:BURS:TRIG:SOUR?",
            set_cmd=f":SOUR{channel}:BURS:TRIG:SOUR {{}}",
            vals=vals.Enum("INT", "EXT", "MAN")
        )

        self.add_parameter(
            "idle",
            get_cmd=f":SOUR{channel}:BURST:IDLE?",
            set_cmd=f":SOUR{channel}:BURST:IDLE {{}}",
            vals=vals.MultiType(
                vals.Enum("FPT", "TOP", "BOTTOM", "CENTER"),
                vals.Numbers()  # DIY
            )
        )

    def trigger(self) -> None:
        """
        Send a software trigger to the instrument. This only works if the
        trigger source is set to manual.
        """
        self.parent.write_raw(f":SOUR{self.channel}:BURS:TRIG")


class DG1062Channel(InstrumentChannel):

    min_impedance = 1
    max_impedance = 10000

    waveform_params = {
        waveform: ["freq", "ampl", "offset", "phase"] for waveform in
        ["HARM", "NOIS", "RAMP", "SIN", "SQU", "TRI", "USER", "PULS"]
    }

    waveform_params["DC"] = ["freq", "ampl", "offset"]
    waveform_params["ARB"] = ["sample_rate", "ampl", "offset"]

    """
    Responses from the machine don't always match
    the name to set the function, hence a translater
    """
    waveform_translate = {"HARM": "HARM", "NOISE": "NOIS",
                          "RAMP": "RAMP", "SIN": "SIN",
                          "SQU": "SQU",   "TRI": "TRI",
                          "USER": "USER", "PULSE": "PULS"}

    waveforms = list(waveform_params.keys())

    def __init__(self, parent: 'DG1062', name: str, channel: int):
        """
        Args:
            parent: The instrument this channel belongs to
            name (str)
            channel (int)
        """

        super().__init__(parent, name)
        self.channel = channel

        for param, unit in [
            ("freq", "Hz"),
            ("ampl", "V"),
            ("offset", "V"),
            ("phase", "deg"),
            ("sample_rate", "1/s")
        ]:
            self.add_parameter(
                param,
                unit=unit,
                get_cmd=partial(self._get_waveform_param, param),
                set_cmd=partial(self._set_waveform_param, param),
            )

        self.add_parameter(
            "waveform",
            get_cmd=partial(self._get_waveform_param, "waveform")
        )

        self.add_parameter(
            "impedance",
            get_cmd=f":OUTPUT{channel}:IMP?",
            set_cmd=f":OUTPUT{channel}:IMP {{}}",
            unit="Ohm",
            vals=vals.MultiType(
                vals.Ints(
                    min_value=DG1062Channel.min_impedance,
                    max_value=DG1062Channel.max_impedance
                ),
                vals.Enum("INF", "MIN", "MAX", "HighZ")
            ),
            get_parser=(lambda value: "HighZ"
                            if float(value) > DG1062Channel.max_impedance
                            else float(value)),
            set_parser=lambda value: "INF" if value == "HighZ" else value
        )

        self.add_parameter(
            "sync",
            get_cmd=f":OUTPUT{channel}:SYNC?",
            set_cmd=f"OUTPUT{channel}:SYNC {{}}",
            vals=vals.Enum(0, 1, "ON", "OFF"),
        )

        self.add_parameter(
            "polarity",
            get_cmd=f":OUTPUT{channel}:GAT:POL?",
            set_cmd=f":OUTPUT{channel}:GAT:POL {{}}",
            vals=vals.OnOff(),
            val_mapping={1: 'POSITIVE', 0: 'NEGATIVE'},
        )

        self.add_parameter(
            "state",
            get_cmd=f"OUTPUT{channel}:STATE?",
            set_cmd=f"OUTPUT{channel}:STATE {{}}",
        )

        self.add_parameter(
            "duty_cycle",
            get_cmd=self._get_duty_cycle,
            set_cmd=self._set_duty_cycle,
            unit="%",
            vals=vals.Numbers(min_value=1, max_value=99),
            docstring=('This functions reads/sets the duty '
                       'cycle for a square and pulse wave '
                       'since these inherit a duty cycle.\n'
                       'For other waveforms it will give '
                       'the user an error')
        )

        burst = DG1062Burst(cast(DG1062, self.parent), "burst", self.channel)
        self.add_submodule("burst", burst)

        # We want to be able to do the following:
        # >>> help(gd.channels[0].sin)
        # >>> gd.channels[0].sin(freq=2E3, ampl=1.0, offset=0, phase=0)
        # We do not use add_function as it is more cumbersome to use.
        for waveform in self.waveforms:
            f = partial_with_docstring(
                self.apply,
                docstring="Args: " + ", ".join(self.waveform_params[waveform]),
                waveform=waveform
            )
            setattr(self, waveform.lower(), f)

        # Retrieve current waveform from device
        self.waveform()

    def apply(self, **kwargs: Any) -> None:
        """
        Public interface to apply a waveform on the channel
        Example:
        >>> gd = DG1062("gd", "TCPIP0::169.254.187.99::inst0::INSTR")
        >>> gd.channels[0].apply(waveform="SIN", freq=1E3, ampl=1.0, offset=0, phase=0)
        Valid waveforms are: HARM, NOIS, RAMP, SIN, SQU, TRI, USER, DC, ARB
        To find the correct arguments of each waveform we can e.g. do:
        >>> help(gd.channels[0].sin)
        Notice the lower case when accessing the waveform through convenience
        functions.
        If not kwargs are given a dictionary with the current waveform
        parameters are returned.
        """
        self._set_waveform_params(**kwargs)

    def current_waveform(self) -> Dict[Any, Any]:
        """Public interface to get the current waveform"""
        return self._get_waveform_params()

    def _get_waveform_param(self, param: str) -> float:
        """
        Get a parameter of the current waveform. Valid param names are
        dependent on the waveform type (e.g. "DC" does not have a "phase")
        """
        params_dict = self._get_waveform_params()
        return params_dict.get(param, None)

    def _get_waveform_params(self) -> Dict[Any, Any]:
        """
        Get all the parameters of the current waveform and
        """
        def to_float(string: str) -> Union[float, str]:
            try:
                return float(string)
            except ValueError:
                return string

        waveform_str = self.parent.ask_raw(f":SOUR{self.channel}:APPL?")
        parts = waveform_str.strip("\"").split(",")

        current_waveform = self.waveform_translate[parts[0]]
        param_vals: List[Union[str, float]] = [current_waveform]
        param_vals += [to_float(i) for i in parts[1:]]
        param_names = ["waveform"] + self.waveform_params[current_waveform]
        params_dict = dict(zip(param_names, param_vals))

        return params_dict

    def _set_waveform_param(self, param: str, value: float) -> None:
        """
        Set a particular waveform param to the given value.
        """
        params_dict = self._get_waveform_params()

        if param in params_dict:
            params_dict[param] = value
        else:
            log.warning(f"Warning, unable to set '{param}' for the current "
                        f"waveform")
            return

        return self._set_waveform_params(**params_dict)

    def _set_waveform_params(self,
                             **params_dict: Union[int, float]) -> None:
        """
        Apply a waveform with values given in a dictionary.
        """
        if not "waveform" in params_dict:
            raise ValueError("At least 'waveform' argument needed")

        waveform = str(params_dict["waveform"])
        if waveform not in self.waveform_params:
            raise ValueError(f"Unknown waveform '{waveform}'. Options are "
                             f"{self.waveform_params.keys()}")

        param_names = self.waveform_params[waveform]

        if not set(param_names).issubset(params_dict.keys()):
            raise ValueError(f"Waveform {waveform} needs at least parameters "
                             f"{param_names}")

        string = f":SOUR{self.channel}:APPL:{waveform} "
        values = [f"{params_dict[param]:7e}" for param in param_names]
        string += ",".join(values)
        self.parent.write_raw(string)

    def _get_duty_cycle(self) -> float:
        """
        Reads the duty cycle after checking waveform
        """
        wf = self.waveform()

        if wf in ['PULS', 'SQU']:
            duty_cycle = self.parent.ask_raw(f":SOUR{self.channel}:FUNC:{wf}:DCYC?")
        else:
            raise ValueError(f"Current function does not contain duty cycle. Current function: {wf}")

        return duty_cycle

    def _set_duty_cycle(self, duty_cycle: float) -> None:
        """
        Sets the duty cycle after checking waveform
        """
        wf = self.waveform()

        if wf in ['PULS', 'SQU']:
            self.parent.write_raw(f":SOUR{self.channel}:FUNC:{wf}:DCYC {duty_cycle}")
        else:
            raise ValueError(f"Current function does not have duty cycle"
                             f" hence can not set. Current function: {wf}")


class DG1062(VisaInstrument):
    """
    Instrument driver for the Rigol DG1062
    """

    waveforms = DG1062Channel.waveforms

    def __init__(self, name: str, address: str,
                 **kwargs: Any):

        super().__init__(name, address, terminator="\n", **kwargs)

        channels = ChannelList(self, "channel", DG1062Channel,
                               snapshotable=False)

        for ch_num in [1, 2]:
            ch_name = f"ch{ch_num}"
            channel = DG1062Channel(self, ch_name, ch_num)
            channels.append(channel)
            self.add_submodule(ch_name, channel)

        channels.lock()
        self.add_submodule("channels", channels)
        self.connect_message()



"""Standard location_provider class(es) for creating DataSet locations."""
from typing import cast
from datetime import datetime
import re
import string

import qcodes

class SafeFormatter(string.Formatter):

    """Modified string formatter that doesn't complain about missing keys."""

    def get_value(self, key, args, kwargs):
        """Missing keys just get left as they were: '{key}'."""
        try:
            return super().get_value(key, args, kwargs)
        except:
            return f'{{{key}}}'


class FormatLocation:

    """
    This is the default DataSet Location provider.

    It provides a callable that returns a new (not used by another DataSet)
    location string, based on a format string ``fmt`` and a dict ``record`` of
    information to pass to ``fmt``.

    Default record items are ``date``, ``time``, and ``counter``
    Record item priority from lowest to highest (double items will be
    overwritten):

    - current ``date``, and ``time``
    - record dict from ``__init__``
    - record dict from ``__call__``
    - automatic ``counter``

    For example if any record dict contains a `date` keyword, it will no longer
    be auto-generated.

    Uses ``io.list`` to search for existing data at a matching location.

    ``counter`` must NOT be provided in the record. If ``fmt`` contains
    '{counter}', we look for existing files matching everything BEFORE this,
    then find the highest counter (integer) among those files and use the next
    value.

    If the format string does not contain ``{counter}`` but the location we
    would return is occupied, we add ``'_{counter}'`` to the end.

    Usage::

        loc_provider = FormatLocation(
            fmt='{date}/#{counter}_{time}_{name}_{label}')
        loc = loc_provider(DiskIO('.'),
                           record={'name': 'Rainbow', 'label': 'test'})
        loc
        > '2016-04-30/#001_13-28-15_Rainbow_test'

    Args:
        fmt (Optional[str]): a format string that all the other info will be
            inserted into. Default '{date}/{time}', or '{date}/{time}_{name}'
            if there is a ``name`` in the record.

        fmt_date (Optional[str]): a ``datetime.strftime`` format string,
            should only use the date part. The result will be inserted in
            '{date}' in ``fmt``. Default '%Y-%m-%d'.

        fmt_time (Optional[str]): a ``datetime.strftime`` format string,
            should only use the time part. The result will be inserted in
            '{time}' in ``fmt``. Default '%H-%M-%S'.

        fmt_counter (Optional[str]): a format string for the counter (integer)
            which is automatically generated from existing DataSets that the
            io manager can see. Default '{03}'.

        record (Optional[dict]): A dict of default values to provide when
            calling the location_provider. Values provided later will
            override these values.

    Note:
        Do not include date/time or number formatting in ``fmt`` itself, such
        as '{date:%Y-%m-%d}' or '{counter:03}'
    """

    default_fmt = qcodes.config['core']['default_fmt']
    default_fmt = cast(str, default_fmt)

    def __init__(self, fmt=None, fmt_date=None, fmt_time=None,
                 fmt_counter=None, record=None):
        # TODO(giulioungaretti) this should be
        # FormatLocation.default_fmt
        self.fmt = fmt or self.default_fmt
        self.fmt_date = fmt_date or '%Y-%m-%d'
        self.fmt_time = fmt_time or '%H-%M-%S'
        self.fmt_counter = fmt_counter or '{:03}'
        self.base_record = record
        self.formatter = SafeFormatter()

        self.counter = 0
        for testval in (1, 23, 456, 7890):
            if self._findint(self.fmt_counter.format(testval)) != testval:
                raise ValueError('fmt_counter must produce a correct integer '
                                 'representation of its argument (eg "{:03}")',
                                 fmt_counter)

    def _findint(self, s):
        try:
            return int(re.findall(r'\d+', s)[0])
        except:
            return 0

    def __call__(self, io, record=None):
        """
        Call the location provider to get a new location.

        Args:
            io (io_manager): where we intend to put the new DataSet.

            record (Optional[dict]): information to insert in the format string
                Any key provided here will override the default record
        """
        loc_fmt = self.fmt

        time_now = datetime.now()
        date = time_now.strftime(self.fmt_date)
        time = time_now.strftime(self.fmt_time)
        format_record = {'date': date, 'time': time}

        if self.base_record:
            format_record.update(self.base_record)
        if record:
            format_record.update(record)

        if 'counter' in format_record:
            raise KeyError('you must not provide a counter in your record.',
                           format_record)

        if ('name' in format_record) and ('{name}' not in loc_fmt):
            loc_fmt += '_{name}'

        if '{counter}' not in loc_fmt:
            location = self.formatter.format(loc_fmt, **format_record)
            if io.list(location):
                loc_fmt += '_{counter}'
                # redirect to the counter block below, but starting from 2
                # because the already existing file counts like 1
                existing_count = 1
            else:
                return location
        else:
            # if counter is already in loc_fmt, start from 1
            existing_count = 0

        # now search existing files for the next allowed counter

        head_fmt = loc_fmt.split('{counter}', 1)[0]
        # io.join will normalize slashes in head to match the locations
        # returned by io.list
        head = io.join(self.formatter.format(head_fmt, **format_record))

        file_list = io.list(head + '*', maxdepth=0, include_dirs=True)

        for f in file_list:
            cnt = self._findint(f[len(head):])
            existing_count = max(existing_count, cnt)

        self.counter = existing_count + 1
        format_record['counter'] = self.fmt_counter.format(self.counter)
        location = self.formatter.format(loc_fmt, **format_record)

        return location



import logging
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

import numpy as np

from qcodes import VisaInstrument
from qcodes.instrument.channel import ChannelList, InstrumentChannel
from qcodes.instrument.parameter import ArrayParameter
from qcodes.utils.validators import ComplexNumbers, Enum, Ints, Numbers

log = logging.getLogger(__name__)


class SR86xBufferReadout(ArrayParameter):
    """
    The parameter array that holds read out data. We need this to be compatible
    with qcodes.Measure

    Args:
        name: Name of the parameter.
        instrument: The instrument to add this parameter to.
    """
    def __init__(self, name: str, instrument: 'SR86x', **kwargs: Any) -> None:

        unit = "deg"
        if name in ["X", "Y", "R"]:
            unit = "V"

        super().__init__(name,
                         shape=(1,),  # dummy initial shape
                         unit=unit,
                         setpoint_names=('Time',),
                         setpoint_labels=('Time',),
                         setpoint_units=('s',),
                         instrument=instrument,
                         docstring='Holds an acquired (part of the) data '
                                   'buffer of one channel.',
                         **kwargs)

        self._capture_data: Optional[np.ndarray] = None

    def prepare_readout(self, capture_data: np.ndarray) -> None:
        """
        Prepare this parameter for readout.

        Args:
            capture_data: The data to capture.
        """
        self._capture_data = capture_data

        data_len = len(capture_data)
        self.shape = (data_len,)
        self.setpoint_units = ('',)
        self.setpoint_names = ('sample_nr',)
        self.setpoint_labels = ('Sample number',)
        self.setpoints = (tuple(np.arange(0, data_len)),)

    def get_raw(self) -> np.ndarray:
        """
        Public method to access the capture data
        """
        if self._capture_data is None:
            raise ValueError(f"Cannot return data for parameter {self.name}. "
                             f"Please prepare for readout by calling "
                             f"'get_capture_data' with appropriate "
                             f"configuration settings")

        return self._capture_data


class SR86xBuffer(InstrumentChannel):
    """
    The buffer module for the SR86x driver. This driver has been verified to
    work with the SR860 and SR865. For reference, please consult the SR860
    manual: http://thinksrs.com/downloads/PDFs/Manuals/SR860m.pdf
    """

    def __init__(self, parent: 'SR86x', name: str) -> None:
        super().__init__(parent, name)
        self._parent = parent

        self.add_parameter(
            "capture_length_in_kb",
            label="get/set capture length",
            get_cmd="CAPTURELEN?",
            set_cmd="CAPTURELEN {}",
            set_parser=self._set_capture_len_parser,
            get_parser=int,
            unit="kB"
        )
        self.bytes_per_sample = 4
        self.min_capture_length_in_kb = 1  # i.e. minimum buffer size
        self.max_capture_length_in_kb = 4096  # i.e. maximum buffer size
        # Maximum amount of kB that can be read per single CAPTUREGET command
        self.max_size_per_reading_in_kb = 64

        self.add_parameter(  # Configure which parameters we want to capture
            "capture_config",
            label="capture configuration",
            get_cmd="CAPTURECFG?",
            set_cmd="CAPTURECFG {}",
            val_mapping={"X": "0", "X,Y": "1", "R,T": "2", "X,Y,R,T": "3"}
        )

        self.add_parameter(
            "capture_rate_max",
            label="capture rate maximum",
            get_cmd="CAPTURERATEMAX?",
            get_parser=float
        )

        self.add_parameter(
            "capture_rate",
            label="capture rate raw",
            get_cmd="CAPTURERATE?",
            set_cmd="CAPTURERATE {}",
            get_parser=float,
            set_parser=self._set_capture_rate_parser
        )

        max_rate = self.capture_rate_max()
        self.available_frequencies = [max_rate / 2 ** i for i in range(20)]

        self.add_parameter(  # Are we capturing at the moment?
            "capture_status",
            label="capture status",
            get_cmd="CAPTURESTAT?"
        )

        self.add_parameter(
            "count_capture_bytes",
            label="captured bytes",
            get_cmd="CAPTUREBYTES?",
            unit="B",
            get_parser=int,
            docstring="Number of bytes captured so far in the buffer. Can be "
                      "used to track live progress."
        )

        self.add_parameter(
            "count_capture_kilobytes",
            label="captured kilobytes",
            get_cmd="CAPTUREPROG?",
            unit="kB",
            docstring="Number of kilobytes captured so far in the buffer, "
                      "rounded-up to 2 kilobyte chunks. Capture must be "
                      "stopped before requesting the value of this "
                      "parameter. If the acquisition wrapped during operating "
                      "in Continuous mode, then the returned value is "
                      "simply equal to the current capture length."
        )

        for parameter_name in ["X", "Y", "R", "T"]:
            self.add_parameter(
                parameter_name,
                parameter_class=SR86xBufferReadout
            )

    def snapshot_base(self, update: Optional[bool] = False,
                      params_to_skip_update: Optional[Sequence[str]] = None
                      ) -> Dict[Any, Any]:
        if params_to_skip_update is None:
            params_to_skip_update = []
        # we omit count_capture_kilobytes from the snapshot because
        # it can only be read after a completed capture and will
        # timeout otherwise when the snapshot is updated, e.g. at
        # station creation time
        params_to_skip_update = list(params_to_skip_update)
        params_to_skip_update.append('count_capture_kilobytes')

        snapshot = super().snapshot_base(update, params_to_skip_update)
        return snapshot

    def _set_capture_len_parser(self, capture_length_in_kb: int) -> int:
        """
        Parse the capture length in kB according to the way buffer treats it
        (refer to the manual for details). The given value has to fit in the
        range and has to be even, otherwise this function raises exceptions.

        Args:
            capture_length_in_kb: The desired capture length in kB.

        Returns:
            capture_length_in_kb
        """
        if capture_length_in_kb % 2:
            raise ValueError("The capture length should be an even number")

        if not self.min_capture_length_in_kb \
                <= capture_length_in_kb \
                <= self.max_capture_length_in_kb:
            raise ValueError(f"The capture length should be between "
                             f"{self.min_capture_length_in_kb} and "
                             f"{self.max_capture_length_in_kb}")

        return capture_length_in_kb

    def set_capture_rate_to_maximum(self) -> None:
        """
        Sets the capture rate to maximum. The maximum capture rate is
        retrieved from the device, and depends on the current value of the
        time constant.
        """
        self.capture_rate(self.capture_rate_max())

    def _set_capture_rate_parser(self, capture_rate_hz: float) -> int:
        """
        According to the manual, the capture rate query returns a value in
        Hz, but then setting this value it is expected to give a value n,
        where the capture rate in Hz is given by
        capture_rate_hz =  max_rate / 2 ** n. Please see page 136 of the
        manual. Here n is an integer in the range [0, 20].

        Args:
            capture_rate_hz: The desired capture rate in Hz. If the desired
                rate is more than 1 Hz from the nearest valid rate, a warning
                is issued and the nearest valid rate it used.

        Returns:
            n_round
        """
        max_rate = self.capture_rate_max()
        n = np.log2(max_rate / capture_rate_hz)
        n_round = int(round(n))

        if not 0 <= n_round <= 20:
            raise ValueError(f"The chosen frequency is invalid. Please "
                             f"consult the SR860 manual at page 136. "
                             f"The maximum capture rate is {max_rate}")

        nearest_valid_rate = max_rate / 2 ** n_round
        if abs(capture_rate_hz - nearest_valid_rate) > 1:
            available_frequencies = ", ".join(
                str(f) for f in self.available_frequencies
            )
            log.warning(f"Warning: Setting capture rate to {nearest_valid_rate:.5} Hz")
            log.warning(f"The available frequencies are: {available_frequencies}")

        return n_round

    def start_capture(self, acquisition_mode: str, trigger_mode: str) -> None:
        """
        Start an acquisition. Please see page 137 of the manual for a detailed
        explanation.

        Args:
            acquisition_mode: "ONE" | "CONT"
            trigger_mode: "IMM" | "TRIG" | "SAMP"
        """

        if acquisition_mode not in ["ONE", "CONT"]:
            raise ValueError(
                "The acquisition mode needs to be either 'ONE' or 'CONT'")

        if trigger_mode not in ["IMM", "TRIG", "SAMP"]:
            raise ValueError(
                "The trigger mode needs to be either 'IMM', 'TRIG' or 'SAMP'")

        cmd_str = f"CAPTURESTART {acquisition_mode}, {trigger_mode}"
        self.write(cmd_str)

    def stop_capture(self) -> None:
        """Stop a capture"""
        self.write("CAPTURESTOP")

    def _get_list_of_capture_variable_names(self) -> List[str]:
        """
        Retrieve the list of names of variables (readouts) that are
        set to be captured
        """
        return self.capture_config().split(",")

    def _get_number_of_capture_variables(self) -> int:
        """
        Retrieve the number of variables (readouts) that are
        set to be captured
        """
        capture_variables = self._get_list_of_capture_variable_names()
        n_variables = len(capture_variables)
        return n_variables

    def _calc_capture_size_in_kb(self, sample_count: int) -> int:
        """
        Given the number of samples to capture, calculate the capture length
        that the buffer needs to be set to in order to fit the requested
        number of samples. Note that the number of activated readouts is
        taken into account.
        """
        n_variables = self._get_number_of_capture_variables()
        total_size_in_kb = int(
            np.ceil(n_variables * sample_count * self.bytes_per_sample / 1024)
        )
        # Make sure that the total size in kb is an even number, as expected by
        # the instrument
        if total_size_in_kb % 2:
            total_size_in_kb += 1
        return total_size_in_kb

    def set_capture_length_to_fit_samples(self, sample_count: int) -> None:
        """
        Set the capture length of the buffer to fit the given number of
        samples.

        Args:
            sample_count: Number of samples that the buffer has to fit
        """
        total_size_in_kb = self._calc_capture_size_in_kb(sample_count)
        self.capture_length_in_kb(total_size_in_kb)

    def wait_until_samples_captured(self, sample_count: int) -> None:
        """
        Wait until the given number of samples is captured. This function
        is blocking and has to be used with caution because it does not have
        a timeout.

        Args:
            sample_count: Number of samples that needs to be captured
        """
        n_captured_bytes = 0
        n_variables = self._get_number_of_capture_variables()
        n_bytes_to_capture = sample_count * n_variables * self.bytes_per_sample
        while n_captured_bytes < n_bytes_to_capture:
            n_captured_bytes = self.count_capture_bytes()

    def get_capture_data(self, sample_count: int) -> Dict[str, np.ndarray]:
        """
        Read the given number of samples of the capture data from the buffer.

        Args:
            sample_count: number of samples to read from the buffer

        Returns:
            The keys in the dictionary correspond to the captured
            variables. For instance, if before the capture, the capture
            config was set as 'capture_config("X,Y")', then the keys will
            be "X" and "Y". The values in the dictionary are numpy arrays
            of numbers.
        """
        total_size_in_kb = self._calc_capture_size_in_kb(sample_count)
        capture_variables = self._get_list_of_capture_variable_names()
        n_variables = self._get_number_of_capture_variables()

        values = self._get_raw_capture_data(total_size_in_kb)

        # Remove zeros which mark the end part of the buffer that is not
        # filled with captured data
        values = values[values != 0]

        values = values.reshape((-1, n_variables)).T
        values = values[:, :sample_count]

        data = {k: v for k, v in zip(capture_variables, values)}

        for capture_variable in capture_variables:
            buffer_parameter = getattr(self, capture_variable)
            buffer_parameter.prepare_readout(data[capture_variable])

        return data

    def _get_raw_capture_data(self, size_in_kb: int) -> np.ndarray:
        """
        Read data from the buffer from its beginning avoiding the instrument
        limit of 64 kilobytes per reading.

        Args:
            size_in_kb :Size of the data that needs to be read; if it exceeds
                the capture length, an exception is raised.

        Returns:
            A one-dimensional numpy array of the requested data. Note that the
            returned array contains data for all the variables that are
            mentioned in the capture config.
        """
        current_capture_length = self.capture_length_in_kb()
        if size_in_kb > current_capture_length:
            raise ValueError(f"The size of the requested data ({size_in_kb}kB) "
                             f"is larger than current capture length of the "
                             f"buffer ({current_capture_length}kB).")

        values = np.array([])
        data_size_to_read_in_kb = size_in_kb
        n_readings = 0

        while data_size_to_read_in_kb > 0:
            offset = n_readings * self.max_size_per_reading_in_kb

            if data_size_to_read_in_kb > self.max_size_per_reading_in_kb:
                size_of_this_reading = self.max_size_per_reading_in_kb
            else:
                size_of_this_reading = data_size_to_read_in_kb

            data_from_this_reading = self._get_raw_capture_data_block(
                size_of_this_reading,
                offset_in_kb=offset)
            values = np.append(values, data_from_this_reading)

            data_size_to_read_in_kb -= size_of_this_reading
            n_readings += 1

        return values

    def _get_raw_capture_data_block(self,
                                    size_in_kb: int,
                                    offset_in_kb: int = 0
                                    ) -> np.ndarray:
        """
        Read data from the buffer. The maximum amount of data that can be
        read with this function (size_in_kb) is 64kB (this limitation comes
        from the instrument). The offset argument can be used to navigate
        along the buffer.

        An exception will be raised if either size_in_kb or offset_in_kb are
        longer that the *current* capture length (number of kB of data that is
        captured so far rounded up to 2kB chunks). If (offset_in_kb +
        size_in_kb) is longer than the *current* capture length,
        the instrument returns the wrapped data.

        For more information, refer to the description of the "CAPTUREGET"
        command in the manual.

        Args:
            size_in_kb: Amount of data in kB that is to be read from the buffer
            offset_in_kb: Offset within the buffer of where to read the data;
                for example, when 0 is specified, the data is read from the
                start of the buffer.

        Returns:
            A one-dimensional numpy array of the requested data. Note that the
            returned array contains data for all the variables that are
            mentioned in the capture config.
        """
        if size_in_kb > self.max_size_per_reading_in_kb:
            raise ValueError(f"The size of the requested data ({size_in_kb}kB) "
                             f"is larger than maximum size that can be read "
                             f"at once ({self.max_size_per_reading_in_kb}kB).")

        # Calculate the size of the data captured so far, in kB, rounded up
        # to 2kB chunks
        size_of_currently_captured_data = int(
            np.ceil(np.ceil(self.count_capture_bytes() / 1024) / 2) * 2
        )

        if size_in_kb > size_of_currently_captured_data:
            raise ValueError(f"The size of the requested data ({size_in_kb}kB) "
                             f"cannot be larger than the size of currently "
                             f"captured data rounded up to 2kB chunks "
                             f"({size_of_currently_captured_data}kB)")

        if offset_in_kb > size_of_currently_captured_data:
            raise ValueError(f"The offset for reading the requested data "
                             f"({offset_in_kb}kB) cannot be larger than the "
                             f"size of currently captured data rounded up to "
                             f"2kB chunks "
                             f"({size_of_currently_captured_data}kB)")

        values = self._parent.visa_handle.query_binary_values(
            f"CAPTUREGET? {offset_in_kb}, {size_in_kb}",
            datatype='f',
            is_big_endian=False,
            expect_termination=False)
        # the sr86x does not include an extra termination char on binary
        # messages so we set expect_termination to False

        return np.array(values)

    def capture_one_sample_per_trigger(
            self,
            trigger_count: int,
            start_triggers_pulsetrain: Callable[..., Any]
    ) -> Dict[str, np.ndarray]:
        """
        Capture one sample per each trigger, and return when the specified
        number of triggers has been received.

        Args:
            trigger_count: Number of triggers to capture samples for
            start_triggers_pulsetrain: By calling this *non-blocking*
                function, the train of trigger pulses should start

        Returns:
            The keys in the dictionary correspond to the captured
            variables. For instance, if before the capture, the capture
            config was set as 'capture_config("X,Y")', then the keys will
            be "X" and "Y". The values in the dictionary are numpy arrays
            of numbers.
        """
        self.set_capture_length_to_fit_samples(trigger_count)
        self.start_capture("ONE", "SAMP")
        start_triggers_pulsetrain()
        self.wait_until_samples_captured(trigger_count)
        self.stop_capture()
        return self.get_capture_data(trigger_count)

    def capture_samples_after_trigger(self,
                                      sample_count: int,
                                      send_trigger: Callable[..., Any]
                                      ) -> Dict[str, np.ndarray]:
        """
        Capture a number of samples after a trigger has been received.
        Please refer to page 135 of the manual for details.

        Args:
            sample_count: Number of samples to capture
            send_trigger: By calling this *non-blocking* function, one trigger
                should be sent that will initiate the capture

        Returns:
            The keys in the dictionary correspond to the captured
            variables. For instance, if before the capture, the capture
            config was set as 'capture_config("X,Y")', then the keys will
            be "X" and "Y". The values in the dictionary are numpy arrays
            of numbers.
        """
        self.set_capture_length_to_fit_samples(sample_count)
        self.start_capture("ONE", "TRIG")
        send_trigger()
        self.wait_until_samples_captured(sample_count)
        self.stop_capture()
        return self.get_capture_data(sample_count)

    def capture_samples(self, sample_count: int) -> Dict[str, np.ndarray]:
        """
        Capture a number of samples at a capture rate, starting immediately.
        Unlike the "continuous" capture mode, here the buffer does not get
        overwritten with the new data once the buffer is full.

        The function blocks until the required number of samples is acquired,
        and returns them.

        Args:
            sample_count: Number of samples to capture

        Returns:
            The keys in the dictionary correspond to the captured
            variables. For instance, if before the capture, the capture
            config was set as 'capture_config("X,Y")', then the keys will
            be "X" and "Y". The values in the dictionary are numpy arrays
            of numbers.
        """
        self.set_capture_length_to_fit_samples(sample_count)
        self.start_capture("ONE", "IMM")
        self.wait_until_samples_captured(sample_count)
        self.stop_capture()
        return self.get_capture_data(sample_count)


class SR86xDataChannel(InstrumentChannel):
    """
    Implements a data channel of SR86x lock-in amplifier. Parameters that are
    assigned to these channels get plotted on the display of the instrument.
    Moreover, there are commands that allow to conveniently retrieve the values
    of the parameters that are currently assigned to the data channels.

    This class relies on the available parameter names that should be
    mentioned in the lock-in amplifier class in `PARAMETER_NAMES` attribute.

    Args:
        parent: an instance of SR86x driver
        name: data channel name that is to be used to reference it from the
            parent
        cmd_id: this ID is used in VISA commands to refer to this data channel,
            usually is an integer number
        cmd_id_name: this name can also be used in VISA commands along with
            channel_id; it is not used in this implementation, but is added
            for reference
        color: every data channel is also referred to by the color with which it
            is being plotted on the instrument's screen; added here only for
            reference
    """
    def __init__(self, parent: 'SR86x', name: str, cmd_id: str,
                 cmd_id_name: Optional[str] = None,
                 color: Optional[str] = None) -> None:
        super().__init__(parent, name)

        self._cmd_id = cmd_id
        self._cmd_id_name = cmd_id_name
        self._color = color

        self.add_parameter(f'assigned_parameter',
                           label=f'Data channel {cmd_id} parameter',
                           docstring=f'Allows to set and get the '
                                     f'parameter that is assigned to data '
                                     f'channel {cmd_id}',
                           set_cmd=f'CDSP {cmd_id}, {{}}',
                           get_cmd=f'CDSP? {cmd_id}',
                           val_mapping=self.parent.PARAMETER_NAMES
                           )

    @property
    def cmd_id(self) -> str:
        return self._cmd_id

    @property
    def cmd_id_name(self) -> Optional[str]:
        return self._cmd_id_name

    @property
    def color(self) -> Optional[str]:
        return self._color


class SR86x(VisaInstrument):
    """
    This is the code for Stanford_SR865 Lock-in Amplifier
    """
    _VOLT_TO_N = {1: 0, 500e-3: 1, 200e-3: 2,
                  100e-3: 3, 50e-3: 4, 20e-3: 5,
                  10e-3: 6, 5e-3: 7, 2e-3: 8,
                  1e-3: 9, 500e-6: 10, 200e-6: 11,
                  100e-6: 12, 50e-6: 13, 20e-6: 14,
                  10e-6: 15, 5e-6: 16, 2e-6: 17,
                  1e-6: 18, 500e-9: 19, 200e-9: 20,
                  100e-9: 21, 50e-9: 22, 20e-9: 23,
                  10e-9: 24, 5e-9: 25, 2e-9: 26,
                  1e-9: 27}
    _N_TO_VOLT = {v: k for k, v in _VOLT_TO_N.items()}

    _CURR_TO_N = {1e-6: 0, 500e-9: 1, 200e-9: 2,
                  100e-9: 3, 50e-9: 4, 20e-9: 5,
                  10e-9: 6, 5e-9: 7, 2e-9: 8,
                  1e-9: 9, 500e-12: 10, 200e-12: 11,
                  100e-12: 12, 50e-12: 13, 20e-12: 14,
                  10e-12: 15, 5e-12: 16, 2e-12: 17,
                  1e-12: 18, 500e-15: 19, 200e-15: 20,
                  100e-15: 21, 50e-15: 22, 20e-15: 23,
                  10e-15: 24, 5e-15: 25, 2e-15: 26,
                  1e-15: 27}
    _N_TO_CURR = {v: k for k, v in _CURR_TO_N.items()}

    _VOLT_ENUM = Enum(*_VOLT_TO_N.keys())
    _CURR_ENUM = Enum(*_CURR_TO_N.keys())

    _INPUT_SIGNAL_TO_N = {
        'voltage': 0,
        'current': 1,
    }
    _N_TO_INPUT_SIGNAL = {v: k for k, v in _INPUT_SIGNAL_TO_N.items()}

    PARAMETER_NAMES = {
                'X': '0',   # X output, 'X'
                'Y': '1',   # Y output, 'Y'
                'R': '2',   # R output, 'R'
                'P': '3',   # theta output, 'THeta'
          'aux_in1': '4',   # Aux In 1, 'IN1'
          'aux_in2': '5',   # Aux In 2, 'IN2'
          'aux_in3': '6',   # Aux In 3, 'IN3'
          'aux_in4': '7',   # Aux In 4, 'IN4'
           'Xnoise': '8',   # X noise, 'XNOise'
           'Ynoise': '9',   # Y noise, 'YNOise'
         'aux_out1': '10',  # Aux Out 1, 'OUT1'
         'aux_out2': '11',  # Aux Out 2, 'OUT2'
            'phase': '12',  # Reference Phase, 'PHAse'
        'amplitude': '13',  # Sine Out Amplitude, 'SAMp'
       'sine_outdc': '14',  # DC Level, 'LEVel'
        'frequency': '15',  # Int. Ref. Frequency, 'FInt'
    'frequency_ext': '16',  # Ext. Ref. Frequency, 'FExt'
    }

    _N_DATA_CHANNELS = 4

    def __init__(
            self,
            name: str,
            address: str,
            max_frequency: float,
            reset: bool = False,
            **kwargs: Any):
        super().__init__(name, address, terminator='\n', **kwargs)
        self._max_frequency = max_frequency
        # Reference commands
        self.add_parameter(name='frequency',
                           label='Frequency',
                           unit='Hz',
                           get_cmd='FREQ?',
                           set_cmd='FREQ {}',
                           get_parser=float,
                           vals=Numbers(
                               min_value=1e-3,
                               max_value=self._max_frequency)
                           )
        self.add_parameter(name='sine_outdc',
                           label='Sine out dc level',
                           unit='V',
                           get_cmd='SOFF?',
                           set_cmd='SOFF {}',
                           get_parser=float,
                           vals=Numbers(min_value=-5, max_value=5))
        self.add_parameter(name='amplitude',
                           label='Amplitude',
                           unit='V',
                           get_cmd='SLVL?',
                           set_cmd='SLVL {}',
                           get_parser=float,
                           vals=Numbers(min_value=0, max_value=2))
        self.add_parameter(name='harmonic',
                           label='Harmonic',
                           get_cmd='HARM?',
                           get_parser=int,
                           set_cmd='HARM {:d}',
                           vals=Ints(min_value=1, max_value=99))
        self.add_parameter(name='phase',
                           label='Phase',
                           unit='deg',
                           get_cmd='PHAS?',
                           set_cmd='PHAS {}',
                           get_parser=float,
                           vals=Numbers(min_value=-3.6e5, max_value=3.6e5))
        # Signal commands
        self.add_parameter(name='sensitivity',
                           label='Sensitivity',
                           get_cmd='SCAL?',
                           set_cmd='SCAL {:d}',
                           get_parser=self._get_sensitivity,
                           set_parser=self._set_sensitivity
                           )
        self.add_parameter(name='filter_slope',
                           label='Filter slope',
                           unit='dB/oct',
                           get_cmd='OFSL?',
                           set_cmd='OFSL {}',
                           val_mapping={6: 0,
                                        12: 1,
                                        18: 2,
                                        24: 3})
        self.add_parameter(name='sync_filter',
                           label='Sync filter',
                           get_cmd='SYNC?',
                           set_cmd='SYNC {}',
                           val_mapping={'OFF': 0,
                                        'ON': 1})
        self.add_parameter(name='noise_bandwidth',
                           label='Noise bandwidth',
                           unit='Hz',
                           get_cmd='ENBW?',
                           get_parser=float)
        self.add_parameter(name='signal_strength',
                           label='Signal strength indicator',
                           get_cmd='ILVL?',
                           get_parser=int)
        self.add_parameter(name='signal_input',
                           label='Signal input',
                           get_cmd='IVMD?',
                           get_parser=self._get_input_config,
                           set_cmd='IVMD {}',
                           set_parser=self._set_input_config,
                           vals=Enum(*self._INPUT_SIGNAL_TO_N.keys()))
        self.add_parameter(name='input_range',
                           label='Input range',
                           unit='V',
                           get_cmd='IRNG?',
                           set_cmd='IRNG {}',
                           val_mapping={1: 0,
                                        300e-3: 1,
                                        100e-3: 2,
                                        30e-3: 3,
                                        10e-3: 4})
        self.add_parameter(name='input_config',
                           label='Input configuration',
                           get_cmd='ISRC?',
                           set_cmd='ISRC {}',
                           val_mapping={'a': 0,
                                        'a-b': 1})
        self.add_parameter(name='input_shield',
                           label='Input shield',
                           get_cmd='IGND?',
                           set_cmd='IGND {}',
                           val_mapping={'float': 0,
                                        'ground': 1})
        self.add_parameter(name='input_gain',
                           label='Input gain',
                           unit='ohm',
                           get_cmd='ICUR?',
                           set_cmd='ICUR {}',
                           val_mapping={1e6: 0,
                                        100e6: 1})
        self.add_parameter(name='adv_filter',
                           label='Advanced filter',
                           get_cmd='ADVFILT?',
                           set_cmd='ADVFILT {}',
                           val_mapping={'OFF': 0,
                                        'ON': 1})
        self.add_parameter(name='input_coupling',
                           label='Input coupling',
                           get_cmd='ICPL?',
                           set_cmd='ICPL {}',
                           val_mapping={'ac': 0, 'dc': 1})
        self.add_parameter(name='time_constant',
                           label='Time constant',
                           unit='s',
                           get_cmd='OFLT?',
                           set_cmd='OFLT {}',
                           val_mapping={1e-6: 0, 3e-6: 1,
                                        10e-6: 2, 30e-6: 3,
                                        100e-6: 4, 300e-6: 5,
                                        1e-3: 6, 3e-3: 7,
                                        10e-3: 8, 30e-3: 9,
                                        100e-3: 10, 300e-3: 11,
                                        1: 12, 3: 13,
                                        10: 14, 30: 15,
                                        100: 16, 300: 17,
                                        1e3: 18, 3e3: 19,
                                        10e3: 20, 30e3: 21})

        self.add_parameter(
            name="external_reference_trigger",
            label="External reference trigger mode",
            get_cmd="RTRG?",
            set_cmd="RTRG {}",
            val_mapping={
                "SIN": 0,
                "POS": 1,
                "POSTTL": 1,
                "NEG": 2,
                "NEGTTL": 2,
            },
            docstring="The triggering mode for synchronization of the "
                      "internal reference signal with the externally provided "
                      "one"
        )

        self.add_parameter(
            name="reference_source",
            label="Reference source",
            get_cmd="RSRC?",
            set_cmd="RSRC {}",
            val_mapping={
                "INT": 0,
                "EXT": 1,
                "DUAL": 2,
                "CHOP": 3
            },
            docstring="The source of the reference signal"
        )

        self.add_parameter(
            name="external_reference_trigger_input_resistance",
            label="External reference trigger input resistance",
            get_cmd="REFZ?",
            set_cmd="REFZ {}",
            val_mapping={
                "50": 0,
                "50OHMS": 0,
                0: 0,
                "1M": 1,
                "1MEG": 1,
                1: 1,
            },
            docstring="Input resistance of the input for the external "
                      "reference signal"
        )

        # Auto functions
        self.add_function('auto_range', call_cmd='ARNG')
        self.add_function('auto_scale', call_cmd='ASCL')
        self.add_function('auto_phase', call_cmd='APHS')

        # Data transfer
        # first 4 parameters from a list of 16 below.
        self.add_parameter('X',
                           label='In-phase Magnitude',
                           get_cmd='OUTP? 0',
                           get_parser=float,
                           unit='V')
        self.add_parameter('Y',
                           label='Out-phase Magnitude',
                           get_cmd='OUTP? 1',
                           get_parser=float,
                           unit='V')
        self.add_parameter('R',
                           label='Magnitude',
                           get_cmd='OUTP? 2',
                           get_parser=float,
                           unit='V')
        self.add_parameter('P',
                           label='Phase',
                           get_cmd='OUTP? 3',
                           get_parser=float,
                           unit='deg')
        self.add_parameter('complex_voltage',
                           label='Voltage',
                           get_cmd=self._get_complex_voltage,
                           unit='V',
                           vals=ComplexNumbers())

        # CH1/CH2 Output Commands
        self.add_parameter('X_offset',
                           label='X offset ',
                           unit='%',
                           get_cmd='COFP? 0',
                           set_cmd='COFP 0, {}',
                           get_parser=float,
                           vals=Numbers(min_value=-999.99, max_value=999.99))
        self.add_parameter('Y_offset',
                           label='Y offset',
                           unit='%',
                           get_cmd='COFP? 1',
                           set_cmd='COFP 1, {}',
                           get_parser=float,
                           vals=Numbers(min_value=-999.99, max_value=999.99))
        self.add_parameter('R_offset',
                           label='R offset',
                           unit='%',
                           get_cmd='COFP? 2',
                           set_cmd='COFP 2, {}',
                           get_parser=float,
                           vals=Numbers(min_value=-999.99, max_value=999.99))
        self.add_parameter('X_expand',
                           label='X expand multiplier',
                           get_cmd='CEXP? 0',
                           set_cmd='CEXP 0, {}',
                           val_mapping={'OFF': '0',
                                        'X10': '1',
                                        'X100': '2'})
        self.add_parameter('Y_expand',
                           label='Y expand multiplier',
                           get_cmd='CEXP? 1',
                           set_cmd='CEXP 1, {}',
                           val_mapping={'OFF': 0,
                                        'X10': 1,
                                        'X100': 2})
        self.add_parameter('R_expand',
                           label='R expand multiplier',
                           get_cmd='CEXP? 2',
                           set_cmd='CEXP 2, {}',
                           val_mapping={'OFF': 0,
                                        'X10': 1,
                                        'X100': 2})

        # Aux input/output
        for i in [0, 1, 2, 3]:
            self.add_parameter(f'aux_in{i}',
                               label=f'Aux input {i}',
                               get_cmd=f'OAUX? {i}',
                               get_parser=float,
                               unit='V')
            self.add_parameter(f'aux_out{i}',
                               label=f'Aux output {i}',
                               get_cmd=f'AUXV? {i}',
                               get_parser=float,
                               set_cmd=f'AUXV {i}, {{}}',
                               unit='V')

        # Data channels:
        # 'DAT1' (green), 'DAT2' (blue), 'DAT3' (yellow), 'DAT4' (orange)
        data_channels = ChannelList(self, "data_channels", SR86xDataChannel,
                                    snapshotable=False)
        for num, color in zip(range(self._N_DATA_CHANNELS),
                              ('green', 'blue', 'yellow', 'orange')):
            cmd_id = f"{num}"
            cmd_id_name = f"DAT{num + 1}"
            ch_name = f"data_channel_{num + 1}"

            data_channel = SR86xDataChannel(
                self, ch_name, cmd_id, cmd_id_name, color)

            data_channels.append(data_channel)
            self.add_submodule(ch_name, data_channel)

        data_channels.lock()
        self.add_submodule("data_channels", data_channels)

        # Interface
        self.add_function('reset', call_cmd='*RST')

        self.add_function('disable_front_panel', call_cmd='OVRM 0')
        self.add_function('enable_front_panel', call_cmd='OVRM 1')

        buffer = SR86xBuffer(self, f"{self.name}_buffer")
        self.add_submodule("buffer", buffer)

        self.input_config()
        self.connect_message()

    def _set_units(self, unit: str) -> None:
        for param in [self.X, self.Y, self.R, self.sensitivity]:
            param.unit = unit

    def _get_complex_voltage(self) -> complex:
        x, y = self.get_values('X', 'Y')
        return x + 1.0j*y

    def _get_input_config(self, s: int) -> str:
        mode = self._N_TO_INPUT_SIGNAL[int(s)]

        if mode == 'voltage':
            self.sensitivity.vals = self._VOLT_ENUM
            self._set_units('V')
        else:
            self.sensitivity.vals = self._CURR_ENUM
            self._set_units('A')

        return mode

    def _set_input_config(self, s: str) -> int:
        if s == 'voltage':
            self.sensitivity.vals = self._VOLT_ENUM
            self._set_units('V')
        else:
            self.sensitivity.vals = self._CURR_ENUM
            self._set_units('A')

        return self._INPUT_SIGNAL_TO_N[s]

    def _get_sensitivity(self, s: int) -> float:
        if self.signal_input() == 'voltage':
            return self._N_TO_VOLT[int(s)]
        else:
            return self._N_TO_CURR[int(s)]

    def _set_sensitivity(self, s: float) -> int:
        if self.signal_input() == 'voltage':
            return self._VOLT_TO_N[s]
        else:
            return self._CURR_TO_N[s]

    def get_values(self, *parameter_names: str) -> Tuple[float, ...]:
        """
        Get values of 2 or 3 parameters that are measured by the lock-in
        amplifier. These values are guaranteed to come from the same
        measurement cycle as opposed to getting values of parameters one by
        one (for example, by calling `sr.X()`, and then `sr.Y()`.

        Args:
            *parameter_names: 2 or 3 names of parameters for which the values
                are requested; valid names can be found in `PARAMETER_NAMES`
                attribute of the driver class

        Returns:
            a tuple of 2 or 3 floating point values

        """
        if not 2 <= len(parameter_names) <= 3:
            raise KeyError(
                'It is only possible to request values of 2 or 3 parameters '
                'at a time.')

        for name in parameter_names:
            if name not in self.PARAMETER_NAMES:
                raise KeyError(f'{name} is not a valid parameter name. Refer '
                               f'to `PARAMETER_NAMES` for a list of valid '
                               f'parameter names')

        p_ids = [self.PARAMETER_NAMES[name] for name in parameter_names]
        output = self.ask(f'SNAP? {",".join(p_ids)}')
        return tuple(float(val) for val in output.split(','))

    def get_data_channels_values(self) -> Tuple[float, ...]:
        """
        Queries the current values of the data channels

        Returns:
            tuple of 4 values of the data channels
        """
        output = self.ask('SNAPD?')
        return tuple(float(val) for val in output.split(','))

    def get_data_channels_parameters(self, query_instrument: bool = True
                                     ) -> Tuple[str, ...]:
        """
        Convenience method to query a list of parameters which the data
        channels are currently assigned to.

        Args:
            query_instrument: If set to False, the internally cashed names of
                the parameters will be returned; if True, then the names will
                be queried through the instrument

        Returns:
            a tuple of 4 strings of parameter names
        """
        if query_instrument:
            method_name = 'get'
        else:
            method_name = 'get_latest'

        return tuple(
            getattr(getattr(self.data_channels[i], 'assigned_parameter'),
                    method_name)()
            for i in range(self._N_DATA_CHANNELS)
        )

    def get_data_channels_dict(self, requery_names: bool = False
                               ) -> Dict[str, float]:
        """
        Returns a dictionary where the keys are parameter names currently
        assigned to the data channels, and values are the values of those
        parameters.

        Args:
            requery_names: if False, the currently assigned parameter names
                will not be queries from the instrument in order to save time
                on communication, in this case the cached assigned parameter
                names will be used for the keys of the dicitonary; if True,
                the assigned parameter names will be queried from the
                instrument

        Returns:
            a dictionary where keys are names of parameters assigned to the
            data channels, and values are the values of those parameters
        """
        parameter_names = self.get_data_channels_parameters(requery_names)
        parameter_values = self.get_data_channels_values()
        return dict(zip(parameter_names, parameter_values))



"""
Tests for `qcodes.utils.plotting`.
"""
import matplotlib

# set matplotlib backend before importing pyplot
matplotlib.use("Agg")

from matplotlib import pyplot as plt
from pytest import fixture

import qcodes
from qcodes.dataset.plotting import plot_by_id
from qcodes.tests.common import default_config

from .dataset_generators import dataset_with_outliers_generator


@fixture(scope='function')
def dataset_with_data_outside_iqr_high_outlier(dataset):
    return dataset_with_outliers_generator(dataset, data_offset=2, low_outlier=-2,
                                           high_outlier=3)


@fixture(scope='function')
def dataset_with_data_outside_iqr_low_outlier(dataset):
    return dataset_with_outliers_generator(dataset, data_offset=2, low_outlier=-2,
                                           high_outlier=3)


@fixture(scope='function')
def dataset_with_outliers(dataset):
    return dataset_with_outliers_generator(dataset, low_outlier=-3,
                                           high_outlier=3,
                                           background_noise=False)


def test_extend(dataset_with_outliers):
    # this one should clipp the upper values
    run_id = dataset_with_outliers.run_id
    _, cb = plot_by_id(run_id, auto_color_scale=False)
    assert cb[0].extend == 'neither'
    _, cb = plot_by_id(run_id, auto_color_scale=True, cutoff_percentile=(0, 0.5))
    assert cb[0].extend == 'min'
    _, cb = plot_by_id(run_id, auto_color_scale=True, cutoff_percentile=(0.5, 0))
    assert cb[0].extend == 'max'
    plt.close()


def test_defaults(dataset_with_outliers):
    run_id = dataset_with_outliers.run_id

    # plot_by_id loads from the database location provided in the qcodes
    # config. But the tests are supposed to run with the standard config.
    # Therefore we need to backup the db location and add it to the default
    # config context.
    db_location = qcodes.config.core.db_location
    with default_config():
        qcodes.config.core.db_location = db_location
        _, cb = plot_by_id(run_id)
        assert cb[0].extend == 'neither'

        qcodes.config.plotting.auto_color_scale.enabled = True

        _, cb = plot_by_id(run_id)
        assert cb[0].extend == 'both'
    plt.close()



import pytest

from qcodes.instrument_drivers.Keysight.KeysightAgilent_33XXX import WaveformGenerator_33XXX
import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'Keysight_33xxx.yaml@sim')


@pytest.fixture(scope='module')
def driver():
    kw_sim = WaveformGenerator_33XXX('kw_sim',
                                      address='GPIB::1::INSTR',
                                      visalib=visalib)
    yield kw_sim

    kw_sim.close()


def test_init(driver):

    idn_dict = driver.IDN()

    assert idn_dict['vendor'] == 'QCoDeS'

    assert driver.model == '33522B'
    assert driver.num_channels == 2


def test_sync(driver):

    assert driver.sync.output() == 'OFF'
    driver.sync.output('ON')
    assert driver.sync.output() == 'ON'

    assert driver.sync.source() == 1
    driver.sync.source(2)
    assert driver.sync.source() == 2
    driver.sync.source(1)
    driver.sync.output('OFF')

def test_channel(driver):
    assert driver.ch1.function_type() == 'SIN'
    driver.ch1.function_type('SQU')
    assert driver.ch1.function_type() == 'SQU'
    driver.ch1.function_type('SIN')


def test_burst(driver):
    assert driver.ch1.burst_ncycles() == 1
    driver.ch1.burst_ncycles(10)
    assert driver.ch1.burst_ncycles() == 10
    driver.ch1.burst_ncycles(1)
    # the following does not actually work because
    # val parser cannot handle INF being returned.
    # not clear if this is a bug or the instrument get
    # set to something else?
    # driver.ch1.burst_ncycles('INF')
    # assert driver.ch1.burst_ncycles() == 'INF'



from typing import Any

from .private.DP8xx import _RigolDP8xx


class RigolDP832(_RigolDP8xx):
    """
    This is the qcodes driver for the Rigol DP832(A) Power Supply
    """
    def __init__(self, name: str, address: str, **kwargs: Any):
        channel_ranges = [(30.0, 3.0), (30.0, 3.0), (5.0, 3.0)]

        ovp_ranges_std = [(0.01, 33.0), (0.01, 33.0), (0.01, 5.5)]
        ocp_ranges_std = [(0.001, 3.3), (0.001, 3.3), (0.001, 3.3)]

        ovp_ranges_precision = [(0.001, 33.0), (0.001, 33.0), (0.001, 5.5)]
        ocp_ranges_precision = [(0.001, 3.3), (0.001, 3.3), (0.001, 3.3)]

        ovp = (ovp_ranges_std, ovp_ranges_precision)
        ocp = (ocp_ranges_std, ocp_ranges_precision)

        super().__init__(name, address, channel_ranges, ovp, ocp, **kwargs)
